# -*- coding: utf-8 -*-
r"""
abstract base class for tests in multilayer test branch
"""
from python3.shell_commands import writelog
from tests.test_multilayers.ianalyze import IAnalyzer
from abc import ABC, abstractmethod
from pathlib import Path
import shutil


class ITest(ABC):
    r"""
    abstract base class for tests.
    """

    def __init__(self, logfile: Path, cleanup: bool = True, **kwargs) -> None:
        r"""
        Initializes test. Each inheriting test class has to call super().__init__ as this initializes the test
        directories

        Args:
            cleanup(bool): boolean flag if the temp-directory after the test shall be deleting. The default value is
            True.
        """
        self.logfile = logfile
        self.tempdir = self._load_tempdir()
        self.cleanup = cleanup

    def _load_tempdir(self) -> Path:
        r"""
        Loads the temporary directory for the test.
        """
        if (Path.cwd() / f'{self.testsignature}_temp_').is_dir():
            # print('WARNING: temp folder for this test already exists. Removing...')
            shutil.rmtree(Path.cwd() / f'{self.testsignature}_temp_')
        l_p = (Path.cwd() / f'{self.testsignature}_temp_')
        l_p.mkdir()
        return l_p

    @property
    @abstractmethod
    def testsignature(self) -> int:
        r"""
        The test signature helps to identify different test and also ensures unique _temp_-folders for different test-
        routines.

        Returns:
            the test signature
        """

    @abstractmethod
    def __call__(self, *args, **kwargs) -> bool:
        r"""
        perform a single test

        Returns:
            the result of the test.
        """

    @abstractmethod
    def __repr__(self) -> str:
        r"""
        Returns:
            the representation of the testing class.
        """

    def cleanup_test(self) -> None:
        r"""
        Any test subclass inheriting from this class has this cleanup method.
        """
        if not self.tempdir.is_dir():
            raise FileNotFoundError('_temp_-folder for this test does not exist.')
        writelog(self.logfile, message='#' * 100)
        writelog(self.logfile, message='clean up test: remove temporary files...')
        shutil.rmtree(self.tempdir)



# -*- coding: utf-8 -*-
r"""
This constants library holds the test signatures. Test signatures have different use cases:

- creating unique _temp_-folders for tests.

- identify test with single integer
"""
# Test Ids Multilayer Test =============================================================================================
MULTI_PRECISIONWITHOUTDAMPING = 1
MULTI_PRECISIONWITHDAMPING = 2
MULTI_DOMAINWALLS_Z = 3


# -*- coding: utf-8 -*-
r"""
This module takes care of writing input files for calculations with the SpinD-code
"""
from typing import TypeVar, List
from pathlib import Path
from python3.shell_commands import adjust_parameter

PathLike = TypeVar("PathLike", str, Path)


class CWriteInput:
    r"""
    General class for writing input files
    """

    def __init__(self, file: PathLike, **kwargs) -> None:
        r"""
        Initializes writing of input file

        Args:
            name(PathLike): name of the created file or path to the file that shall be created.
        """
        self.file = file
        self.content = dict(**kwargs)

    def __call__(self) -> None:
        r"""
        Calls the writing
        """
        for (key, value) in self.content.items():
            with open(self.file, 'a') as f:
                f.write(key + ' ' + value + '\n')

    def appendline(self, key: str, value: str) -> None:
        r"""
        Appends a line to the input file
        """
        with open(self.file, 'a') as f:
            f.write(key + ' ' + value + '\n')

    def appendmultilinevalues(self, key: str, values: List[str]) -> None:
        r"""
        Some input files have multiple lines of values after a key line. E. g. lattice in lattice in. In this case
        provide the key line as key and each following line in a list of strings
        """
        with open(self.file, 'a') as f:
            f.write(key + '\n')
            for line in values:
                f.write(line + '\n')

    def replaceline(self, key: str, value: str) -> None:
        r"""
        Replaces the line addressed by key with key + value
        """
        adjust_parameter(keyword=key, value=value, directory_file=self.file)



import numpy as np
from visualizations import write_STM
from shell_commands import *
import os
import time
from magnetisations import SpinLattice


def start_spinD(sp, calc_directory, input_directory, dt, duration):
    if not os.path.isdir(calc_directory) :
        make_directory(calc_directory)
    copy_element(input_directory + "/inp", calc_directory)
    copy_element(input_directory + "/lattice.in", calc_directory)
    copy_element(input_directory + "/efield.in", calc_directory)
    copy_element(input_directory + "/dyna.in", calc_directory)
    copy_element(input_directory + "/job.sh", calc_directory)
    copy_element(input_directory + "/simu.in", calc_directory)
    adjust_parameter("timestep", dt, calc_directory + "/dyna.in")
    adjust_parameter("duration", duration, calc_directory + "/dyna.in")
    adjust_parameter("Nsize", " " +str(sp.size) + " " + str(sp.size) + " 1", calc_directory + "/lattice.in")
    # set simulation
    adjust_parameter("i_metropolis", '.F.', calc_directory + "/simu.in")
    adjust_parameter("i_paratemp", '.F.', calc_directory + "/simu.in")
    adjust_parameter("i_sd", '.T.', calc_directory + "/simu.in")
    adjust_parameter("i_gneb", '.F.', calc_directory + "/simu.in")
    adjust_parameter("i_htst", '.F.', calc_directory + "/simu.in")
    adjust_parameter("i_vt", '.F.', calc_directory + "/simu.in")
    adjust_parameter("i_min", '.F.', calc_directory + "/simu.in")
    with change_directory(calc_directory):
        write_STM(sp)
        print 'Spin Dynamic at: ', calc_directory
        if cluster() :
            adjust_parameter('#SBATCH --job-name=', '#SBATCH --job-name=spinD', 'job.sh')
            call_sbatch("job.sh")
        else :
            call_spin()



def start_minimisation(sp, calc_directory, input_directory):
    if not os.path.isdir(calc_directory) :
        make_directory(calc_directory)
    copy_element(input_directory + "/inp", calc_directory)
    copy_element(input_directory + "/lattice.in", calc_directory)
    copy_element(input_directory + "/efield.in", calc_directory)
    #copy_element(input_directory + "/dyna.in", calc_directory)
    copy_element(input_directory + "/job.sh", calc_directory)
    copy_element(input_directory + "/simu.in", calc_directory)
    copy_element(input_directory + "/minimization.in", calc_directory)
    adjust_parameter("Nsize", " " +str(sp.size) + " " + str(sp.size) + " 1", calc_directory + "/lattice.in")
    # set simulation
    adjust_parameter("i_metropolis", '.F.', calc_directory + "/simu.in")
    adjust_parameter("i_paratemp", '.F.', calc_directory + "/simu.in")
    adjust_parameter("i_sd", '.F.', calc_directory + "/simu.in")
    adjust_parameter("i_gneb", '.F.', calc_directory + "/simu.in")
    adjust_parameter("i_htst", '.F.', calc_directory + "/simu.in")
    adjust_parameter("i_vt", '.F.', calc_directory + "/simu.in")
    adjust_parameter("i_min", '.T.', calc_directory + "/simu.in")
    with change_directory(calc_directory):
        write_STM(sp)
        print 'Minimisation at: ', calc_directory
        if cluster() :
            adjust_parameter('#SBATCH --job-name=', '#SBATCH --job-name=spinD', 'job.sh')
            call_sbatch("job.sh")
        else :
            call_spin()


#===============================================================================
# INPUT: spinlattice object @sl
#        directory for calculation, string: @calcdirectory
#        directory for the input files, string: @input_directory
#OUTPUT: 1d array with energy contributions from all interactions
#===============================================================================
def numeric_energy(sl, calc_directory, input_directory):
    # SUBFUNCTION: read out energy_density_end.dat
    def get_calenergy_data(calc_directory):
        EXC = []
        DMI = []
        ANI = []
        ZEE = []
        DIP = []
        BIQ = []
        FSP = []
        TOT = []
        for n in range(images) :
            exc = dmi = ani = zee = dip = biq = fsp = tot = 0.0
            enerden = calc_directory + '/energy_density_end.dat'
            with open(enerden) as file:
                for line in file:
                    data = np.asarray(line.split())
                    if data[0][0] == '#':
                        continue
                    exc += float(data[2])
                    dmi += float(data[3])
                    ani += float(data[4])
                    zee += float(data[5])
                    dip += float(data[6])
                    biq += float(data[7])
                    fsp += float(data[8])
                    tot += float(data[9])
            EXC.append(exc)
            DMI.append(dmi)
            ANI.append(ani)
            ZEE.append(zee)
            DIP.append(dip)
            BIQ.append(biq)
            FSP.append(fsp)
            TOT.append(tot)
        return EXC, DMI, ANI, ZEE, DIP, BIQ, FSP, TOT
    # start spindynamics for just 1 step with zero time evolution
    dt = 0
    duration=1
    adjust_parameter("CalEnergy", '.T.', input_directory + "/inp")
    start_spinD(sl, calc_directory, input_directory, dt, duration)
    # extract energies
    while not os.path.isfile(calc_directory + '/tot_energy_end.dat'):
        time.sleep(0.2)
    data = np.loadtxt(calc_directory + '/tot_energy_end.dat', unpack=True)
    Exc_intra = data[2]
    Exc_inter = data[3]
    Dmi = data[4]
    Ani = data[5]
    Zee = data[6]
    Dip = data[7]
    Biq = data[8]
    Sp4 = data[9]
    Sp3 = data[10]
    Abs = data[11]
    return np.asarray([Abs, Exc_intra, Zee, Ani, Dmi, Sp3, Sp4, Biq, Dip])



def update_lattice(calc_directory):
    SpinSTM_end = calc_directory + '/SpinSTM_end.dat'
    while not os.path.isfile(SpinSTM_end):
        time.sleep(1)
    sp = SpinLattice(path=SpinSTM_end)
    return sp



def start_GNEB(sp_init, sp_final, calc_directory, input_directory, images, iterations, use_path=None, HTST=True):
    if not os.path.isfile(calc_directory) :
        make_directory(calc_directory)
    copy_element(input_directory + "/inp", calc_directory)
    copy_element(input_directory + "/lattice.in", calc_directory)
    copy_element(input_directory + "/efield.in", calc_directory)
    copy_element(input_directory + "/job.sh", calc_directory)
    copy_element(input_directory + "/GNEB.in", calc_directory)
    copy_element(input_directory + "/simu.in", calc_directory)
    adjust_parameter("i_metropolis", '.F.', calc_directory + "/simu.in")
    adjust_parameter("i_paratemp", '.F.', calc_directory + "/simu.in")
    adjust_parameter("i_sd", '.F.', calc_directory + "/simu.in")
    adjust_parameter("i_gneb", '.T.', calc_directory + "/simu.in")
    adjust_parameter("i_vt", '.F.', calc_directory + "/simu.in")
    adjust_parameter("i_min", '.F.', calc_directory + "/simu.in")
    if HTST :
        copy_element(input_directory + "/htst.in", calc_directory)
        adjust_parameter("i_htst", '.T.', calc_directory + "/simu.in")
    adjust_parameter("nim", images, calc_directory + "/GNEB.in")
    adjust_parameter("mep_itrmax", iterations, calc_directory + "/GNEB.in")
    if sp_init.size == sp_final.size :
        adjust_parameter("Nsize", " " +str(sp_init.size) + " " + str(sp_init.size) + " 1", calc_directory + "/lattice.in")
    else :
        print 'ERROR: Initial and Final configurations have different dimensions'
        quit()
    if use_path :
        copy_all_elements(use_path, '.dat', calc_directory)
        adjust_parameter("amp_rnd_path", 0.0, calc_directory + "/GNEB.in")
    with change_directory(calc_directory):
        write_STM(sp_init, name='momfile_i')
        write_STM(sp_final, name='momfile_f')
        print 'GNEB at: ', calc_directory
        if cluster() :
            adjust_parameter('#SBATCH --job-name=', '#SBATCH --job-name=GNEB', 'job.sh')
            call_sbatch("job.sh")
        else :
            call_spin()





def start_VecTrans(SL, mode, steps, calc_directory, input_directory, jobname='vt'):
    if not os.path.isfile(calc_directory) :
        make_directory(calc_directory)
    copy_element(input_directory + "/inp", calc_directory)
    copy_element(input_directory + "/lattice.in", calc_directory)
    copy_element(input_directory + "/job.sh", calc_directory)
    copy_element(input_directory + "/vt.in", calc_directory)
    copy_element(input_directory + "/simu.in", calc_directory)
    # adjust input
    adjust_parameter("i_metropolis", '.F.', calc_directory + "/simu.in")
    adjust_parameter("i_paratemp", '.F.', calc_directory + "/simu.in")
    adjust_parameter("i_sd", '.F.', calc_directory + "/simu.in")
    adjust_parameter("i_gneb", '.F.', calc_directory + "/simu.in")
    adjust_parameter("i_htst", '.F.', calc_directory + "/simu.in")
    adjust_parameter("i_vt", '.T.', calc_directory + "/simu.in")
    adjust_parameter("i_min", '.F.', calc_directory + "/simu.in")
    adjust_parameter("vec_init_index", mode, calc_directory + "/vt.in")
    adjust_parameter("translation_steps", steps, calc_directory + "/vt.in")

    with change_directory(calc_directory):
        write_STM(SL, name='SpinSTMi.dat')

        print 'VecTrans at: ', calc_directory
        if cluster() :
            set_jobname(jobname, 'job.sh')
            call_sbatch("job.sh")
        else :
            call_spin()



import numpy as np
import sys
from python3.shell_commands import *

class InputPhys():

    def __init__(self, system=None, file=None):
        if file:
            self.load_inp(file)
        elif system == 'PdFeIr_fcc':
            self.load_PdFeIr()
        else:
            self.load_empty()


    def load_PdFeIr(self):
        self.exc = [14.4046e-3,
                    -2.48108e-3,
                    -2.68507e-3,
                    0.520605e-3,
                    0.73757e-3,
                    0.277615e-3,
                    0.160881e-3,
                    -0.57445e-3,
                    -0.212654e-3]
        self.dmi = [1.0e-3]
        self.ani = [-0.7e-3]
        self.sp3 = [0.0]
        self.sp4 = [0.0]
        self.biq = [0.0]

    def load_empty(self):
        self.exc = [0]
        self.dmi = [0]
        self.ani = [0]
        self.sp3 = [0]
        self.sp4 = [0]
        self.biq = [0]


    def load_inp(self, file):
        self.exc = []
        self.dmi = []
        self.sp3 = []
        self.sp4 = []
        self.biq = []
        self.ani = []
        with open(file, 'r') as f :
            while True:
                buf = f.readline()
                if not buf:
                    break
                words = buf.split()
                for word in words:

                    # check magnetic field
                    if word == 'H_ext': self.H_ext = [float(words[1]), float(words[2]), float(words[3])]
                    elif word == 'J_1' : self.exc.append(float(words[1][:-3])*1e-3)
                    elif word == 'J_2' : self.exc.append(float(words[1][:-3])*1e-3)
                    elif word == 'J_3' : self.exc.append(float(words[1][:-3])*1e-3)
                    elif word == 'J_4' : self.exc.append(float(words[1][:-3])*1e-3)
                    elif word == 'J_5' : self.exc.append(float(words[1][:-3])*1e-3)
                    elif word == 'J_6' : self.exc.append(float(words[1][:-3])*1e-3)
                    elif word == 'J_7' : self.exc.append(float(words[1][:-3])*1e-3)
                    elif word == 'J_8' : self.exc.append(float(words[1][:-3])*1e-3)
                    elif word == 'J_9' : self.exc.append(float(words[1][:-3])*1e-3)
                    elif word == 'J_10' : self.exc.append(float(words[1][:-3])*1e-3)
                    elif word == 'J_11' : self.exc.append(float(words[1][:-3])*1e-3)
                    elif word == 'J_12' : self.exc.append(float(words[1][:-3])*1e-3)
                    elif word == 'J_B' : self.biq.append(float(words[1][:-3])*1e-3)
                    elif word == 'K_1' : self.sp4.append(float(words[1][:-3])*1e-3)
                    elif word == 'Y_1' : self.sp3.append(float(words[1][:-3])*1e-3)
                    elif word == 'D_ani' : self.ani.append(float(words[1][:-3])*1e-3)
                    # check DMI constants
                    elif word == 'D_1' : self.dmi.append(float(words[1][:-3])*1e-3)
                    elif word == 'D_2' : self.dmi.append(float(words[1][:-3])*1e-3)
                    elif word == 'D_3' : self.dmi.append(float(words[1][:-3])*1e-3)
                    elif word == 'D_4' : self.dmi.append(float(words[1][:-3])*1e-3)
                    elif word == 'D_5' : self.dmi.append(float(words[1][:-3])*1e-3)
                    elif word == 'D_6' : self.dmi.append(float(words[1][:-3])*1e-3)
                    elif word == 'D_7' : self.dmi.append(float(words[1][:-3])*1e-3)
                    elif word == 'D_8' : self.dmi.append(float(words[1][:-3])*1e-3)
                    elif word == 'D_9' : self.dmi.append(float(words[1][:-3])*1e-3)
                    elif word == 'D_10' : self.dmi.append(float(words[1][:-3])*1e-3)
                    elif word == 'D_11' : self.dmi.append(float(words[1][:-3])*1e-3)



import numpy as np
import sys


#============================================
# function for the topological charge of a lattice, see B.Berg and M.Luescher
# works just on a lattice constructed in the way
# the class 'Lattices' does. For other constructions
# S1,S2,S3,S4 have to be assigned disfferently.
# But the calculation of the spherical area is nice in every way.
# Input : mag = [my, mx,mz]
# Output : float Q
#============================================
def topocharge(mag):
    Q = 0
    size = int(np.sqrt(len(mag[0])))
    for n in range(size-1):
        for m in range(size-1):
            index1 = n*size + m
            index2 = (n+1)*(size-1) + (m+1) +n
            index3 = (n+1)*(size-1) + (m+2) +n
            index4 = n*size + (m+1)
            S1 = np.asarray([mag[0][index1], mag[1][index1], mag[2][index1]])
            S2 = np.asarray([mag[0][index2], mag[1][index2], mag[2][index2]])
            S3 = np.asarray([mag[0][index3], mag[1][index3], mag[2][index3]])
            S4 = np.asarray([mag[0][index4], mag[1][index4], mag[2][index4]])
            area1 = get_spherical_area(S1, S2, S3)
            area2 = get_spherical_area(S1, S3, S4)
            Q += area1 + area2
    return Q/4.0/np.pi
#============================================
# function returns the spherical area created by three spins
# the sign of the area is the sign of the spat product S1*(S2xS3)
# inspired by implementation of B. Dupe
#============================================
def get_spherical_area(S1, S2, S3):
    con = 1 + np.dot(S1,S2) + np.dot(S2,S3) + np.dot(S3,S1)
    # FM CASE
    if con >= 4.-10e-8 :
        area = 0.
    else :
        mat = [S1, S2, S3]
        det_mat = np.linalg.det(mat)
        # EXCEPTIONAL CASE
        if abs(con) <= 10e-8 and abs(det_mat) <=10e-8:
            area = 0.
        # NON-EXCEPTIONAL NON-COLINEAR CASE
        else :
            area = 2.0*np.arctan2(det_mat, con)
    return area


#============================================
# funktion for calculation of the skyrmion radius.
# method adopted from S.v. Malottki
# Input: SpinLattice object
#============================================
def skyrmionradius(splat):
    from scipy.optimize import curve_fit, fmin
    #============================================
    # 2d fit function for the skyrmion profile
    #============================================
    def sk_2dprofile(R, r0x, r0y, c, w):
        r = np.sqrt((R[0]-r0x)**2 + (R[1]-r0y)**2)
        mzs = np.cos(theta(r,c,w))
        return mzs
    #============================================
    # analytic function for the azimuthal angle @theta of a skyrmion
    # and its derivation @dtheta_dr
    # @r is the distance from the centrum
    # @c, @w the fit parameters
    #============================================
    def theta(r,c,w):
        comp1 = np.arcsin(np.tanh((-abs(r) -c)*2.0/w))
        comp2 = np.arcsin(np.tanh((-abs(r) +c)*2.0/w))
        return np.pi +comp1 +comp2
    def dtheta_dr(r,c,w):
        comp1 = 2.0*np.sqrt(-np.tanh(2.0*(-c + abs(r))/w)**2 + 1)/w
        comp2 = 2.0*np.sqrt(-np.tanh(2.0*( c + abs(r))/w)**2 + 1)/w
        return -comp1 -comp2
    #============================================
    # calculates the radius determined by @c, @w using
    # lilley criteria
    #============================================
    def sk_radius(c,w):
        x0 = fmin(lambda x: dtheta_dr(x,c,w), 0,disp=False,xtol=1.0E-10)[0]
        rad = x0 - theta(x0,c,w)/dtheta_dr(x0,c,w)
        return rad
    #================================================================
    # main
    #================================================================
    # 1) Find the minimum of magnetisation, as it is nearly the sk center
    index = np.where( splat.mz == min(splat.mz))[0][0]
    # 2) make a 2d fit to mz using sk_2dprofile function
    start_para = [splat.X[index], splat.Y[index], 2.5, 3.0]
    popt, pcov = curve_fit(sk_2dprofile, [splat.X, splat.Y], splat.mz, start_para, maxfev=2000)
    popt, pcov = curve_fit(sk_2dprofile, [splat.X, splat.Y], splat.mz, popt)
    # the estimated parameters c,w determine the Radius
    radius = sk_radius(popt[2], popt[3])
    return radius, popt, np.sqrt(np.diag(pcov))

#============================================
# function to find the width of a domainwall using curve_fit
#============================================
def dwwidth(splat):
    from scipy.optimize import curve_fit, fmin
    X, Y, mx, my, mz = np.asarray(splat.X), np.asarray(splat.Y), np.asarray(splat.mx), np.asarray(splat.my), np.asarray(splat.mz)
    #============================================
    # 2d fit function for the domainwall profile and a straight line
    #============================================
    def dw_2dprofile(R, r0, phi, w):
        theta = 2.0*np.arctan(np.exp( (R[0]*np.cos(phi) + R[1]*np.sin(phi) -r0)/w))
        return np.cos(theta)
    def dw_2dprofile_reduzed(R, w):
        theta = 2.0*np.arctan(np.exp( (R[0]*np.cos(init_phi) + R[1]*np.sin(init_phi)-init_r0)/w))
        return np.cos(theta)
    def gerade(x, y0, m):
        return y0 + x*m
    #============================================
    # the functions returns the initial values for @phi, the angle between @direction
    # and the x-axis, and also @r0, the distance to the wall on this line
    #============================================
    def get_init_values():
        # find points inside the DW and fit them with a straight line
        index = np.where( abs(mz) < 0.1 )
        popt, pcov = curve_fit(gerade, X[index[0]], Y[index[0]])
        # calculate the initial guess for the angle between @direction and x-axis
        # by using the slope of the fitted straight line
        phi = np.pi/2.0 - np.arctan(abs(popt[1]))
        # to find @r0 set arg=(r-r0)n=0 => r0=rx*cos(phi) + ry*sin(phi)
        # which corresponds to a position on the domainwall
        r0 = splat.X[index[0][0]]*np.cos(phi) + splat.Y[index[0][0]]*np.sin(phi)
        return r0, phi
    #============================================
    # main
    #============================================
    # determine situation: down->up, up->down
    dir = np.sign(splat.mz[0])
    # get initial values for @r0 and @phi and fit DW with them to get @init_w
    init_r0, init_phi = get_init_values()
    popt, pcov = curve_fit(dw_2dprofile_reduzed, [splat.X, splat.Y], np.asarray(splat.mz)*dir, [1.0])
    init_w = popt[0]
    start_para = [init_r0, init_phi, init_w]
    # fit the DW profile to the data using reasonable start parameters
    popt, pcov = curve_fit(dw_2dprofile, [splat.X, splat.Y], np.asarray(splat.mz)*dir, start_para)
    return popt[0], popt[1], popt[2]








#==============================================
# INPUT: initial and final images @init and @final. optionaly the number of images
#        to generate
# OUTPUT: the methods writes the sequence of images to the folder @'geodasic_path'
# DESCRIPTION: This method calculates @steps images on the geodasic
#              path between the initial image @init and the final image
#              @final.
# CONVENTION: the initial image is @image_0 and the final @image_(steps)
#==============================================
def geodasic_path(init, final, steps=10, make_pov=False):
    import os
    from magnetisations import rotate_to_axis, rotation_matrix, rotate_spins, SpinLattice
    from visualizations import write_STM, spinpov
    from shell_commands import make_directory, change_directory

    assert init.size == final.size
    if not os.path.isdir('geodasic_path'):
        make_directory('geodasic_path')

    with change_directory('geodasic_path'):
        # write the inital and final STM files
        write_STM(init, name='image_0.dat')
        write_STM(final, name='image_'+str(steps)+'.dat')
        if make_pov :
            spinpov(init, 'image_0')
            spinpov(final, 'image_'+str(steps))

        # for each image on the geodasic path, calculate the spin configuration
        for n in range(1,steps):
            sl = SpinLattice(size=init.size, key='hex', magmom=init.magmom[0])
            for m in range(len(init.mx)):
                axis, ang = rotate_to_axis([final.mx[m], final.my[m], final.mz[m]], [init.mx[m], init.my[m], init.mz[m]])
                mat = rotation_matrix(axis, ang*float(n)/float(steps))
                tempx, tempy, tempz = rotate_spins([init.mx[m]], [init.my[m]], [init.mz[m]], mat)
                sl.mx[m], sl.my[m], sl.mz[m] = tempx[0], tempy[0], tempz[0]
            tag = 'image_'+str(n)
            write_STM(sl, name=tag+'.dat')
            if make_pov :
                spinpov(sl, tag)




#==============================================
# INPUT: @SPinLattice OBJECT TO WORK ON
#        @key TO SPECIFY LATTICE TYPE ('hex', OR 'quad')
#        @mode = 'translation' OR 'helicity' TO SPECIFY TYPE OF MODE
# OUTPUT: PARTITION FUNCTION CONSIDERING STATISTICAL VOLUME FOR ZERO-MODE
# DESCRIPTION: METHOD RETURNS THE GOLDSTONE-VOLUME OF A SPECIFIED CONTINUES MODE
#==============================================
def zero_mode_partitionfunction(SL, key=None, mode=None, resolution=1e4):
    # DECIDE WHICH KIND OF MODE WE WANT TO HAVE THE VOLUME FOR
    if mode == None:
        print('ERROR: set @mode= \'translation\' or \'helicity\' ' )
        exit()
    # =================================================
    # TRANSLATION MODE PART
    # =================================================
    elif mode == 'translation':
        if key == None:
            print('ERROR: set @key= \'hex\' or \'quad\' ' )
            exit()
        # WE NEED THE 2D-VOLUME OF THE UNITCELL
        elif key == 'hex':
            trans_vec = np.array([ [0.5, 0.86602540378, 0.0], [1., 0., 0.]])
        elif key == 'quad':
            trans_vec = np.array([ [1., 0., 0.], [0., 1., 0.]])
        unitcell_span = np.cross(trans_vec[0], trans_vec[1])
        unitcell_volume = abs(unitcell_span[2])

        # TRANSLATION ALONG LATTICE VECTOR [0.5, 0.86602540378, 0.0]
        mx_tmp = [None for _ in range(len(SL.mx))]
        my_tmp = [None for _ in range(len(SL.my))]
        mz_tmp = [None for _ in range(len(SL.mz))]
        for n in range(len(SL.mx)) :
            # NEW INDEX FOR TRANSLATED SPIN @n IS FOUND CONSIDERING PERIODIC BOUNDARY CONDITIONS
            if (n+1)%SL.size != 0 :
                new_index = n+1
            else :
                new_index = (n//SL.size)*SL.size
            # COPY CURRENT SPIN TO NEW LOCATION ON TEMPORARY LATTICE @SL_new1
            mx_tmp[new_index], my_tmp[new_index], mz_tmp[new_index] = SL.mx[n], SL.my[n], SL.mz[n]
        # GET THE GEODASIC DISTANCE OF LATTICES
        geo_dist1 = geodasic_dist_data_orientated(SL.mx, SL.my, SL.mz, mx_tmp, my_tmp, mz_tmp)

        # TRANSLATION ALONG LATTICE VECTOR [1.0, 0.0, 0.0]
        for n in range(len(SL.mx)) :
            # NEW INDEX FOR TRANSLATED SPIN @n IS FOUND CONSIDERING PERIODIC BOUNDARY CONDITIONS
            if n < SL.size*(SL.size-1) :
                if (n+1)%SL.size != 0 :
                    new_index = n + SL.size +1
                else :
                    new_index = n+1
            else :
                if (n+1)%SL.size != 0 :
                    new_index = n%SL.size +1
                else :
                    new_index = 0
            # COPY CURRENT SPIN TO NEW LOCATION ON TEMPORARY LATTICE @SL_new2
            mx_tmp[new_index], my_tmp[new_index], mz_tmp[new_index] = SL.mx[n], SL.my[n], SL.mz[n]
        # GET THE GEODASIC DISTANCE OF LATTICES
        geo_dist2 = geodasic_dist_data_orientated(SL.mx, SL.my, SL.mz, mx_tmp, my_tmp, mz_tmp)
        return (geo_dist1*geo_dist2)*unitcell_volume

    # =================================================
    # HELICITY MODE PART
    # =================================================
    elif mode == 'helicity_discrete':
        # DEFINE z-AXIS AS AXIS OF ROTATION AND RESOLUTION FOR POLAR ANGLE @ang
        axis , ang = [0., 0., 1.], 2.*np.pi/resolution
        rot_mat = rotation_matrix(axis, ang)
        # CREATE NEW LATTICE WITH INPLANE COMPONENTS ROTATED BY @ang
        mx_tmp, my_tmp, mz_tmp = rotate_spins(SL.mx, SL.my, SL.mz, rot_mat)
        # SINCE ROTATIONS ARE UNITARY PROJECTIONS, A FULL ROTATION BY 2pi CAN BE INTERPOLATED BY
        # MULTIPLYING THE RESULT BY @resolution
        geo_dist = geodasic_dist_data_orientated(SL.mx, SL.my, SL.mz, mx_tmp, my_tmp, mz_tmp)
        return resolution*geo_dist

    # =================================================
    # SEMI ANALYTICAL APPROACH TO HELICITY VOLUME
    # =================================================
    elif mode == 'helicity':
        dist = 0.
        for n in range(len(SL.mz)):
            dist += 1.-SL.mz[n]**2.
        return 2.*np.pi*np.sqrt(dist)



# =================================================
# FUNCTIONS FOR CALCULATING THE GEODASIC DISTANCE.
# INPUT CAN BE EITHER SPINLATTICE OBJECT OR ARRAYS
# CONTAINING SPINS FOR DATAORIENTATED APPROACH
# =================================================
def geodasic_dist_data_orientated(mx1, my1, mz1, mx2, my2, mz2):
    dist = 0.
    for n in range(len(mx1)):
        vec1 = np.array([mx1[n], my1[n], mz1[n]])
        vec2 = np.array([mx2[n], my2[n], mz2[n]])
        if vec2[0] == None:
            print('NoneType in vec2 at indices: ', n)
            exit()
        prod = np.dot(vec1,vec2)
        norm = np.linalg.norm(np.cross(vec1, vec2))
        if prod >= 1.0:
            continue
        else :
            dist += np.arctan2(norm, prod)**2
    return np.sqrt(dist)


def geodasic_dist(SL1, SL2):
    length = len(SL1.mx)
    dist = 0.
    for n in range(length):
        vec1 = np.array([SL1.mx[n], SL1.my[n], SL1.mz[n]])
        vec2 = np.array([SL2.mx[n], SL2.my[n], SL2.mz[n]])
        prod = np.dot(vec1,vec2)
        norm = np.linalg.norm(np.cross(vec1, vec2))
        if prod >= 1.0:
            continue
        else :
            dist += np.arctan2(norm, prod)**2
    return np.sqrt(dist)




#================================================================
# family of functions to rotate or mirror spin orientations on the lattice
#================================================================
# rodriguez rotation matrix for given angle
#================================================================
def rotation_matrix(axis, angle):
    import scipy.linalg as sclin
    norm = np.linalg.norm(axis)
    if norm > 0:
        return sclin.expm(np.cross(np.eye(3), np.asarray(axis)/norm*angle))
    else :
        return sclin.expm(np.cross(np.eye(3), np.asarray([0.0, 0.0, 0.0])))
#================================================================
# use householder transformation for mirroring vector at plane
# decribed by normal vector plane_normal
#================================================================
def mirror_matrix(plane_normal):
    N = np.asarray(plane_normal)/np.linalg.norm(plane_normal)
    return np.identity(3) - 2.0*np.outer(N,N)
#================================================================
# rotatation matrix to rotate vec to axis
#================================================================
def rotate_to_axis(axis, vec):
    N = np.cross(vec, axis)
    if np.linalg.norm(N) > 0 :
        N = N/np.linalg.norm(N)
        prod = np.dot(np.asarray(axis)/np.linalg.norm(axis),np.asarray(vec)/np.linalg.norm(vec))
        if prod >= 1.0:
            ang = 0.0
        else :
            ang = np.arccos(prod)
    else:
        ang = 0.0
    return N, ang
#================================================================
# rotates all vectors in given set using Mat
#================================================================
def rotate_spins(mx,my,mz, Mat):
    mx_tmp = [None for _ in range(len(mx))]
    my_tmp = [None for _ in range(len(my))]
    mz_tmp = [None for _ in range(len(mz))]
    for n in range(len(mz)):
        mx_tmp[n], my_tmp[n], mz_tmp[n] = np.dot(Mat,np.asarray([mx[n], my[n], mz[n]]))
    return mx_tmp, my_tmp, mz_tmp




#================================================================
# RETURNS THE RECIPROCE REPRESENTATION OF @SpinLattice ON BRAVAIS LATTICE
#================================================================
def bravais_lattice(sp, key='hex'):

    def lattice_fft(self, sp):
        L = sp.size
        xfft = (np.fft.ifft2(np.asarray(sp.mx).reshape(L, L)).transpose()).reshape(sp.size**2)
        yfft = (np.fft.ifft2(np.asarray(sp.my).reshape(L, L)).transpose()).reshape(sp.size**2)
        zfft = (np.fft.ifft2(np.asarray(sp.mz).reshape(L, L)).transpose()).reshape(sp.size**2)
        return xfft, yfft, zfft

    def inverse_lattice_fft(self, fft_x, fft_y, fft_z):
        L = int(np.sqrt(len(fft_x)))
        x = (np.fft.fft2(np.asarray(fft_x).reshape(L, L)).transpose()).reshape(L**2)
        y = (np.fft.fft2(np.asarray(fft_y).reshape(L, L)).transpose()).reshape(L**2)
        z = (np.fft.fft2(np.asarray(fft_z).reshape(L, L)).transpose()).reshape(L**2)
        return x, y, z

    def reciproce_lattice(sp):
        latvec = np.asarray([ [0.5, -0.86602540378, 0.0], [0.5, 0.86602540378, 0.0], [0.0, 0.0, 1.0] ])
        spat = np.dot(latvec[0], np.cross(latvec[1], latvec[2]))
        recvec = np.asarray([ np.cross(latvec[1], latvec[2]), np.cross(latvec[2], latvec[0]), np.cross(latvec[0], latvec[1]) ])*2.0*np.pi/spat
        recX, recY, recZ = [], [], []
        for n in range(sp.size):
            r0 = n*recvec[0]/np.linalg.norm(recvec[0])/sp.size*2
            for m in range(sp.size) :
                r = r0 + m*recvec[1]/np.linalg.norm(recvec[1])/sp.size*2
                recX.append(r[0])
                recY.append(r[1])
                recZ.append(r[2])
        # resize recX, recY to the form of the brillouin zone
        recX = [ (-1.0 +2.0*(x-min(recX))/(max(recX)-min(recX))) for x in recX ]
        recY = [ np.cos(np.pi/6.0)*0.666*(-1.0 +2.0*(y-min(recY))/(max(recY)-min(recY))) for y in recY ]
        return recX, recY, recZ

    # main call
    xfft, yfft, zfft = lattice_fft(sp)
    # get the reciproke coordinates
    recX, recY, recZ = reciproce_lattice(sp)
    return recX, recY, recZ, xfft, yfft, zfft



#==============================================
# AFM FUNCTIONS

#==============================================
#Function to get two sub lattices
#==============================================
def get_AFM_sublattices(sl):
    X1,Y1,Z1,mx1,my1,mz1,magmom1 = [],[],[],[],[],[],[]
    X2,Y2,Z2,mx2,my2,mz2,magmom2 = [],[],[],[],[],[],[]
    #seperate every second spin and write it in to a sublattice
    for n in range(len(sl.X)):
        if (n%2 + n//sl.size)%2 == 0:
            X1.append(sl.X[n])
            Y1.append(sl.Y[n])
            Z1.append(sl.Z[n])
            mx1.append(sl.mx[n])
            my1.append(sl.my[n])
            mz1.append(sl.mz[n])
            magmom1.append(sl.magmom[n])
        else:
            X2.append(sl.X[n])
            Y2.append(sl.Y[n])
            Z2.append(sl.Z[n])
            mx2.append(sl.mx[n])
            my2.append(sl.my[n])
            mz2.append(sl.mz[n])
            mom2.append(sl.magmom[n])
    #make the sublattices a property of the object self
    SubSL1 = [X1, Y1, Z1, mx1 ,my1, mz1, magmom1]
    SubSL2 = [X2, Y2, Z2, mx2, my2, mz2, magmom2]
    return SubSL1, SubSL2



import numpy as np
import sys
sys.path.append('../python3')
from python3.magnetisations import *
from python3.visualizations import *






def run_example_setup_skyrmion_get_radius():
    # GENERAL SETTINGS
    size, key, magmom = 20, 'hex', 1.
    # DEFINE RADIAL THETA-PROFILE PARAMETER
    c, w = 2., 3.
    # DEFINE SKYRMION PARAMETER
    helicity, vorticity = 0, 1.
    center = [size//2, 0]
    # SETUP LATTICE
    SL = SpinLattice(size=size, key=key, path=None, magmom=magmom)
    SL.addSkyrmion(center, vorticity, helicity , c, w)
    SL.getSkradius()
    print('radius   R_sk =', SL.sk_radius)
    write_STM(SL)
    show_lattice(SL, details='Sk')
    display_plots()


def run_example_charge_and_chargedensity():
    # GENERAL SETTINGS
    size, key, magmom = 30, 'hex', 1.
    # DEFINE VORTICITIES m1,m2. TOPOLOGICAL CHARGE IS -Q=m1+m2
    m1, m2 = -1, 1
    center_skyrmion1, center_skyrmion2 = [size//2 , size//4], [ size//2 , -(size//4)]
    # ADD TWO NEEL SKYRMIONS WITH VORTICITIES m1= -6 AND m2=5
    SL = SpinLattice(size=size , key=key , path=None, magmom=magmom)
    SL.addSkyrmion(center_skyrmion1 , m1, np. pi , 3. , 3)
    SL.addSkyrmion(center_skyrmion2 , m2, np. pi , 3. , 3)
    # GET TOPOLOGICAL CHARGE AND PRINT OUT
    SL.getTopocharge()
    print('Tolologic charge: Q =', SL.topo_charge)
    # SHOW LATTICE AND TOPOLOGICAL DENSITY
    show_lattice(SL, details=None)
    show_topodensity(SL)
    display_plots()




def run_example_setup_cycloidal_1qstates():
    # GENERAL SETTINGS
    size, key, magmom = 30, 'hex', 1.
    # INTERNAL COORDINATE SYSTEM
    #
    #           ----------
    #          /         /  \
    #         / (1,0,0)/     \
    #        /       /        \
    #        \       \       /
    #         \(0,1,0) \    /
    #          \         \ /
    #           ---------
    #
    # @Rq and @Iq span the plane, the spins rotate in. given in internal coordinates
    Rq = [0.0, 0.0, 1.0]
    Iq = [1.0, 1.0, 0.0]
    q = 0.05
    Q = np.asarray([ 1.0, 1.0, 0.0])*q
    # CREATE LATTICE OBJECT AND ADD 1q STATE
    sl = SpinLattice(size=size, key=key, path=None, magmom=magmom)
    sl.add1qstate(Rq, Iq, Q)
    # SHOW LATTICE INCLUDING DETAILS FOR SPINSPIRALS
    show_lattice(sl, details='SS')
    display_plots()



def run_example_setup_domainwall_get_width():
    # GENERAL SETTINGS
    size, key, magmom = 40, 'hex', 1.
    # DW settings
    sigma = 2.5
    r0 = size//3
    direction = np.array([0.5, np.sqrt(3.)/6.])
    # create a right rotating neel wall
    SL = SpinLattice(size=size, key=key, path=None, magmom=magmom)
    SL.addDomainwall( r0, direction, sigma, heli=np.pi/2.)
    # GET THE WIDTH FROM FITTING
    SL.getDWwidth()
    print('fitted width:  w =', SL.dw_width)
    show_lattice(SL, details='DW')
    display_plots()



def run_example_setup_double_domainwall():
    # GENERAL SETTINGS
    size, key, magmom = 40, 'hex', 1.
    # DW settings
    direction = np.array([0.5, np.sqrt(3.)/6.])
    sigma = 1.5
    r0_1 = size//5
    r0_2 = 2*size//3
    # create a right rotating neel wall
    SL = SpinLattice(size=size, key=key, path=None, magmom=magmom)
    SL.addDomainwall(r0_1, -direction, sigma, heli=np.pi/2)
    SL.addDomainwall(r0_2, direction, sigma, heli=np.pi)
    show_lattice(SL, details=None)
    display_plots()




def run_example_setup_multilayer_hex():
    c, w = 2., 3.
    helicity, vorticity = 0., 1.
    size, key, magmom1 = 70, 'hex', 1.
    # center is the same for both Skyrmions
    center = [size//2, 0]
    a1_0=2.0/3.0
    a2_0=1.0/3.0
    magmom2= 2.0
    ML =  MultiLayer(number_layers=2)
    # No offset for this layer
    SL0 = SpinLattice(size=size, key=key, path=None, magmom=magmom1)
    SL0.addSkyrmion(center, vorticity, helicity, c, w)
    # Offset for this layer
    SL1 = SpinLattice(size=size, key=key, path=None, magmom=magmom2, a1_0=a1_0, a2_0=a2_0)
    SL1.addSkyrmion(center, vorticity, helicity, c, w)
    ML.add_layer(SL0, index_layer=0, zpos_uc=0.)
    ML.add_layer(SL1, index_layer=1, zpos_uc=3.8302600704)
    # WRITE SpinSTM FOR MULTILAYER
    ML.write_STM(name='ML_hex_SpinSTMi.dat')



def run_example_setup_multilayer_quad():
    c, w = 2., 3.
    helicity, vorticity = 0., 1.
    size, key, magmom1 = 50, 'quad', 1.
    # center is the same for both Skyrmions
    center = [size//2, size//2]
    a1_0=0.5
    a2_0=0.5
    magmom2= 2.0
    ML =  MultiLayer(number_layers=2)
    # No offset for this layer
    SL0 = SpinLattice(size=size, key=key, path=None, magmom=magmom1)
    SL0.addSkyrmion(center, vorticity, helicity, c, w)
    # Offset for this layer
    SL1 = SpinLattice(size=size, key=key, path=None, magmom=magmom2, a1_0=a1_0, a2_0=a2_0)
    SL1.addSkyrmion(center, vorticity, helicity, c, w)
    ML.add_layer(SL0, index_layer=0, zpos_uc=0.)
    ML.add_layer(SL1, index_layer=1, zpos_uc=3.8302600704)
    # WRITE SpinSTM FOR MULTILAYER
    ML.write_STM(name='ML_quad_SpinSTMi.dat')



def run_example_setup_bimeron():
    a1, a2 = 1.0, 1.0
    R = 4
    size, key, magmom1 = 50, 'hex', 1.
    # center is the same for both Skyrmions
    center = [size//2, 0]
    ML =  MultiLayer(number_layers=1)
    SL = SpinLattice(size=size, key=key, path=None, magmom=magmom1)
    SL.add_bimeron(center, 4, 1, 0.1)
    ML.add_layer(SL, index_layer=0, zpos_uc=0.)
    ML.write_STM(name='Bimeron_SpinSTMi.dat')
    show_lattice(SL, details=None)



def run_example_setup_nanoskyrmionlattice():
    # CREATE LATTICE
    size, key, magmom = 27, 'hex', 1.0
    sl = SpinLattice(size=size, key=key, path=None, magmom=magmom)
    # DEFINE RECIPROKE VECTORS FOR PLANES OF ROTATION IN CARTESIAN COORDINATES
    Rq = np.array([0.0, 0.0, 1.0])
    Iq1 = np.array([1., 0., 0.])
    Iq2 = np.array([np.cos((2./3.)*np.pi), -np.sin((2./3.)*np.pi),0])
    Iq3 = np.array([np.cos((2./3.)*np.pi),np.sin((2./3.)*np.pi),0])
    # DEFINE LENGTH OF RECIPROCAL VECTOR @q = 2pi/lambda
    q = 2.*np.pi/4.5
    # CHANGE EVERY SPIN VECTOR BY HAND TO RECIEVE NANOSKYRMION LATTICE
    for n in range(size**2):
        rr = np.array([sl.X[n], sl.Y[n], sl.Z[n]]) + [0, 1./np.sqrt(3.), 0]
        ss1 = 2.*( Rq*np.cos(np.dot(-q*Iq1, rr)) - Iq1*np.sin(np.dot(-q*Iq1, rr)))
        ss2 = 2.*( Rq*np.cos(np.dot(-q*Iq2, rr)) - Iq2*np.sin(np.dot(-q*Iq2, rr)))
        ss3 = 2.*( Rq*np.cos(np.dot(-q*Iq3, rr)) - Iq3*np.sin(np.dot(-q*Iq3, rr)))
        spin = ss1 +ss2 +ss3
        norm = np.sqrt(spin[0]**2. +spin[1]**2. +spin[2]**2.)
        sl.mx[n], sl.my[n], sl.mz[n] = spin[0]/norm, spin[1]/norm, spin[2]/norm
    show_lattice(sl)



def run_example_uudd_by_superposition():
    # SPECIFY SPINSPIRALS
    Rq = [0.0, 0.0, 1.0]
    Iq = [1.0, 0.5, 0.0]
    q = 0.25
    Q =  np.asarray([ 1.0, 0.0, 0.0])
    # SPECIFY LATTICE
    size, key, magmom = 30, 'hex', 1.
    # CREATE STATES WITH LEFT AND RIGHT ROTATING SPINSPIRALS + PHASE SHIFT
    sl1 = SpinLattice(size=size, key=key, path=None, magmom=magmom)
    sl1.add1qstate(Rq, Iq, Q*q, phase=-np.pi/4.)
    sl2 = SpinLattice(size=size, key=key, path=None, magmom=magmom)
    sl2.add1qstate(Rq, Iq, -Q*q, phase=0)
    # CREATE SUPERPOSITION OF BOTH STATES
    sl3 = SpinLattice(size=size, key=key, path=None, magmom=magmom)
    sl3.add_superposition(sl1, sl2)
    # SHOW RESULTING LATTICE
    show_lattice(sl3, filename='uudd.png')



def run_example_setup_AFM_skyrmion():
    #SPECIFY LATTICE
    size, key, magmom = 30, 'quad', 1.
    sl = SpinLattice(size=size, key=key, path=None, magmom=magmom, AFM=True)
    # DEFINE RADIAL THETA-PROFILE PARAMETER
    c, w = 6., 5.
    # DEFINE SKYRMION PARAMETER
    helicity, vorticity = 0, -3.
    center = [size//2, size//2]
    sl.addSkyrmion(center, vorticity, helicity , c, w, AFM=True)
    # DEVIDE LATTICE INTO SUBLATTICES
    sublat1, sublat2 = afm_sublattices(sl)
    # CHECK TOPOLOGICAL CHARGES
    sl.getTopocharge()
    sublat1.getTopocharge()
    sublat2.getTopocharge()
    print('topological charges of lattices and sublattices:')
    print('0) AFM lattice  : ', sl.topo_charge)
    print('1) sub lattice 1: ', sublat1.topo_charge)
    print('2) sub lattice 2: ', sublat2.topo_charge)
    # SHOW RESULTING LATTICE
    show_lattice(sl, filename='afm_skyrmion.png')
    show_lattice(sublat1)
    show_lattice(sublat2)



def run_example_setup_chimera_state():
    #SPECIFY LATTICE
    size, key, magmom = 30, 'hex', 1.
    sl = SpinLattice(size=size, key=key, path=None, magmom=magmom, AFM=False)
    # SPECIFY SKYRMION PROPERTIES
    # SKYRMION SETTINGS
    vorticity = 1.
    helicity = np.pi/2.
    c, w = 3.1, 5.0
    ursprung = [size/2., 0.]
    # WRITE CHIMERA STATE
    sl.add_chimera(ursprung, vorticity, helicity , c, w, sym_chimera=False, angl_chimera=np.pi/2.)
    show_lattice(sl, filename='chimera_state.png', zoom=[sl.size/2, 0, 10])





import subprocess
import os
import sys
from contextlib import contextmanager
import time
from tempfile import mkstemp
from shutil import move, copy2
from os import fdopen, remove


#======================================================================
# Lib for python - shell communication

# Moritz Goerzen
# 08.2018
# Spintronic Theory Kiel
#======================================================================









@contextmanager
def change_directory(newdir):
    #==================================================================
    # changes the directory in the following way:
    #
    # with change_directory(path_string):
    #    print os.getcwd()               # example line
    #
    prevdir = os.getcwd()
    os.chdir(os.path.expanduser(newdir))
    try:
        yield
    finally:
        os.chdir(prevdir)



def make_directory(directory):
    #==================================================================
    bashCommand = "mkdir " + str(directory)
    process = subprocess.Popen(bashCommand.split(), stdout=subprocess.PIPE)
    output, error = process.communicate()



def copy_folder(target, destination):
    #==================================================================
    bashCommand = "cp -r " + str(target) + " " + str(destination)
    process = subprocess.Popen(bashCommand.split(), stdout=subprocess.PIPE)
    output, error = process.communicate()




def copy_element(target, destination):
    #==================================================================
    bashCommand = "cp " + str(target) + " " + str(destination)
    process = subprocess.Popen(bashCommand.split(), stdout=subprocess.PIPE)
    output, error = process.communicate()


def copy_all_elements(targetdirectory, extension, destination):
    #==================================================================
    for filename in os.listdir(targetdirectory):
        if filename.endswith(extension):
            pathname = os.path.join(targetdirectory, filename)
            if os.path.isfile(pathname):
                copy2(pathname, destination)



def remove_element(filename):
    #==================================================================
    bashCommand = "rm " + str(filename)
    process = subprocess.Popen(bashCommand.split(), stdout=subprocess.PIPE)
    output, error = process.communicate()



def remove_folder(directory):
    #==================================================================
    bashCommand = "rm -r " + str(directory)
    process = subprocess.Popen(bashCommand.split(), stdout=subprocess.PIPE)
    output, error = process.communicate()




def call_bash(scriptname):
    #==================================================================
    # calls an .sh-script on linux desktop
    bashCommand = "bash " + str(scriptname)
    process = subprocess.Popen(bashCommand.split(), stdout=subprocess.PIPE)
    output, error = process.communicate()



def call_sbatch(scriptname):
    #==================================================================
    # calls an .sh-script on linux clusters
    bashCommand = "sbatch " + str(scriptname)
    process = subprocess.Popen(bashCommand.split(), stdout=subprocess.PIPE)
    output, error = process.communicate()



def call_python(scriptname):
    #==================================================================
    # calls an external python script
    bashCommand = "python " + str(scriptname)
    process = subprocess.Popen(bashCommand.split(), stdout=subprocess.PIPE)
    output, error = process.communicate()



def adjust_parameter(keyword, value, directory_file):
    #==================================================================
    # looks up "directory_file" for the given "keyword" and substitutes everything behind it
    # with "value"
    fh, abs_path = mkstemp()
    with fdopen(fh,'w') as new_file:
        with open(directory_file) as old_file:
            for line in old_file:
                words = line.split()
                if not words :
                    pass
                elif words[0] == keyword :
                    new_file.write(line.replace(line, keyword + " " + str(value) + "\n"))
                else :
                    new_file.write(line)
    #Remove original file
    remove(directory_file)
    #Move new file
    move(abs_path, directory_file)


def set_jobname(name, directory_file):
    fh, abs_path = mkstemp()
    with fdopen(fh,'w') as new_file:
        with open(directory_file) as file:
            for line in file :
                print 'new line'
                if line.startswith('#SBATCH'):
                    print 'starts with #SBATCH'
                    if line.split()[1].startswith('--job-name=') :
                        print 'is the correct line'
                        new_file.write(line.replace(line, '#SBATCH --job-name='+ name + "\n"))
                    else :
                        new_file.write(line)
                else :
                    new_file.write(line)
    #Remove original file
    remove(directory_file)
    #Move new file
    move(abs_path, directory_file)






def adjust_line_under_key(keyword, value, directory_file):
    #==================================================================
    # looks up "directory_file" for the given "keyword" and substitutes everything behind it
    # with "value"
    fh, abs_path = mkstemp()
    with fdopen(fh,'w') as new_file:
        with open(directory_file) as old_file:
            for line in old_file:
                words = line.split()
                if not words :
                    pass
                elif words[0] == keyword :
                    new_file.write(line)
                    line2 = next(old_file)
                    new_file.write(line2.replace(line2, str(value) + "\n"))
                else :
                    new_file.write(line)
    #Remove original file
    remove(directory_file)
    #Move new file
    move(abs_path, directory_file)




def adjust_parameter_new_version(keyword, value, directory_file):
   #==================================================================
   # looks up "directory_file" for the given "keyword" and substitutes everything behind it
   # with "value"
   keyword_array=keyword.split()
   fh, abs_path = mkstemp()
   with fdopen(fh,'w') as new_file:
       with open(directory_file) as old_file:
           for line in old_file:
               words = line.split()
               # keywords can consist of more than one word. Every word has to match the corresponding element of the current line
               keyword_match=True
               set_space=True
               if len(keyword_array)==0 or len(keyword_array)>len(words):
                  keyword_match=False
               else:
                 for i in range(0,len(keyword_array)):
                    if keyword_array[i] not in words[i]:
                       keyword_match=False
                    elif keyword_array [i] != words [i]:
                       set_space=False
               if not words :
                   pass
               elif keyword_match==True :
                   if set_space==True:
                       new_file.write(line.replace(line, keyword + " " + str(value) + "\n"))
                   else:
                       new_file.write(line.replace(line, keyword + str(value) + "\n"))
               else :
                   new_file.write(line)
   #Remove original file
   remove(directory_file)
   #Move new file
   move(abs_path, directory_file)





def make_file(filename):
    #==================================================================
    # simple touch
    bashCommand = "touch " + str(filename)
    process = subprocess.Popen(bashCommand.split(), stdout=subprocess.PIPE)
    output, error = process.communicate()



def call_spin():
    #==================================================================
    # call SpinD_Kiel code lying on given directory
    #bashCommand = r" ~/code/SpinD/SpinD_Kiel/spin"
    bashCommand = r" ~/code/spind_kiel_v032020/spin"
    #bashCommand = r" ~/code/Sd_solver_HOI_gneb/spin"
    #subprocess.call(bashCommand, shell=True)
    process = subprocess.Popen(bashCommand.split(), stdout=subprocess.PIPE, shell=True)
    output, error = process.communicate()

def call_spin_():
    #==================================================================
    # call SpinD_Kiel code lying on given directory
    bashCommand = r" ~/code/SpinD/SpinD_Kiel/spin"
    #subprocess.call(bashCommand, shell=True)
    process = subprocess.Popen(bashCommand.split(), stdout=subprocess.PIPE, shell=True)
    output, error = process.communicate()



def backuping_mkdir(directory):
    #==================================================================
    # copies folder to oldVersion if already exists and then creates new one as required
    if os.path.isdir(directory) :
        temp_directory = directory + "_backup"
        i = 1
        while os.path.isdir(temp_directory) :
            temp_directory = directory + "_backup" + str(i)
            i += 1
        print 'directory ' + '\"' + directory + '\" ' + 'already exists : backuping old version to ' + '\"' + temp_directory + '\"'
        copy_folder(directory, temp_directory)
        remove_folder(directory)
    make_directory(directory)
    time.sleep(1.0)



def backuping_touch(filename):
    #==================================================================
    # copies file to oldVersion if already exists and then creates new one as required
    if Path(filename).is_file() :
        temp_file = filename + "_backup"
        i = 1
        while Path(temp_file).is_file() :
            temp_file = filename + "_backup" + str(i)
            i += 1
        print 'file ' + '\"' + filename + '\" ' + 'already exists : backuping old version to ' + '\"' + temp_file + '\"'
        copy_element(filename, temp_file)
        remove_element(filename)
    make_file(filename)
    time.sleep(1.0)


def cluster():
    #==================================================================
    # reads the current machines name and returns the bool variable @cluster
    bashCommand = 'hostname'
    process = subprocess.Popen(bashCommand.split(), stdout=subprocess.PIPE)
    output, error = process.communicate()
    if output[0:2]=='rz':
        return True
    elif output[0:3]=='sh8' or output[0:7] == 'goerzen' or output[0:3] == 'cau':
        return False
    else:
        print 'Name =', output
        print 'ERROR at determining current machine. \n Please go to "shell_commands.py" and edit @cluster with the correct machine names '
        quit()
    return cl



def read_last_line(filename):
    #==================================================================
    num_lines = sum(1 for line in open(filename))
    with open(filename, 'rb') as f:
        if num_lines > 1 :
            f.seek(-2, os.SEEK_END)
            while f.read(1) != b'\n':
                f.seek(-2, os.SEEK_CUR)
            return f.readline().decode()
        else :
            return f.readline().decode()



import numpy as np
import os

#============================================
# function to show all created plots
#============================================
def display_plots():
    import matplotlib.pyplot as plt
    plt.show()
#============================================
# write spin configuration
# default name is 'SpinSTMi.dat'
#============================================
def write_STM(sp, name='SpinSTMi.dat'):
    with open(name, 'w') as file :
        for n in range(len(sp.X)):
            file.write('%.8f' %sp.X[n] + '\t'
                    + '%.8f' %sp.Y[n] +  '\t'
                    + '%.8f' %sp.Z[n] +  '\t'
                    + '%.8f' %sp.mx[n] + '\t'
                    + '%.8f' %sp.my[n] + '\t'
                    + '%.8f' %sp.mz[n] + '\t'
                    + '%.8f' %sp.magmom[n] + '\n')


#=============================================
# function to visualize the created lattice with matplotlib
#=============================================
def show_lattice(spinlattice, details=None, filename='spinorientation.png', zoom=None):
    import matplotlib.pyplot as plt
    import matplotlib.colors as mcolors
    from scipy.interpolate import griddata
    from mpl_toolkits.axes_grid1.inset_locator import (inset_axes, InsetPosition, mark_inset)
    #================================================================
    # function adds the parameters that define the DW to the plot
    #================================================================
    def DWdetails(ax):
        direction = np.absolute(spinlattice.direction)
        r0 = spinlattice.r0
        R0 = r0*direction
        size = spinlattice.size
        width = spinlattice.dw_width
        ax.arrow(0, 0, direction[0]*size, direction[1]*size, head_width=1, head_length=1, color='k', ls='-', width=0.00002)
        ax.arrow(R0[0], R0[1], -direction[1]*size/2.0, direction[0]*size/2.0, head_width=0, head_length=0, color='k', ls='-', width=0.00002)
        ax.arrow(R0[0], R0[1], direction[1]*size/2.0, -direction[0]*size/2.0, head_width=0, head_length=0, color='k', ls='-', width=0.00002)
        ax.plot(R0[0], R0[1] , marker='h', markersize=10, c='k', ls='')
        def dw_1dprofile(R, r0, w):
            theta = 2.0*np.arctan(np.exp((R -r0)/w))
            return np.cos(theta)
        axx = plt.axes([0,0,1,1])
        ip = InsetPosition(ax, [0.75,0.75,0.25,0.25])
        axx.set_axes_locator(ip)
        x = np.linspace(0, size, 100)
        axx.plot(x, dw_1dprofile(x, r0, width), ls='-', color='b')
        axx.set_xlabel(r'r|$_{||n}$ /a', fontsize=10)
        axx.set_ylabel(r'mz', fontsize=10)
    #================================================================
    # function adds a circle according to skyrmion radius to the plot
    #================================================================
    def Skdetails(ax):
        popt = spinlattice.sk_popt
        rad = spinlattice.sk_radius
        size = spinlattice.size
        circle = plt.Circle((popt[0], popt[1]), rad, color='k', fill=False)
        ax.add_artist(circle)
        def theta_1d(r,r0,c,w):
            comp1 = np.arcsin(np.tanh((-abs(r-r0) -c)*2.0/w))
            comp2 = np.arcsin(np.tanh((-abs(r-r0) +c)*2.0/w))
            return np.pi +comp1 +comp2
        #axx = plt.axes([0,0,1,1])
        #ip = InsetPosition(ax, [0.75,0.75,0.25,0.25])
        #axx.set_axes_locator(ip)
        #x = np.linspace(0, size, 100)
        #axx.plot(x, theta_1d(x, popt[0], popt[2], popt[3]), ls='-', color='b')
        #axx.set_xlabel(r'x /a', fontsize=10)
        #axx.set_ylabel(r'mz', fontsize=10)
        #ax.set_xlim(size/2 -20, size/2 +20 )
        #ax.set_ylim(-20, 20 )
    #================================================================
    # function adds a direction of @Q to the plot, as well as an inset with
    # the rotation along @Q
    #================================================================
    def SSdetails(ax):
        size = spinlattice.size
        Q = spinlattice.Q
        ax.arrow(0, 0, Q[0]*size, Q[1]*size, head_width=1, head_length=1, color='k', ls='-', width=0.00002)
    #================================================================
    # FUNCTION FOR CREATION OF USER DEFINED COLORMAP
    # THANKS TO stackoverflow!
    #================================================================
    def user_defined_colormap():
        c = mcolors.ColorConverter().to_rgb
        c1 = (240./255., 80./255., 0.)
        c2 = (0., 0., 168./255.)
        c3 = (81./255., 163./255, 54./255.)
        rvb = make_colormap([c1, c2, 0.5, c2, c3])
        #rvb = make_colormap([c('red'), c('violet'), 0.5, c('violet'), c('blue')])
        return rvb

    def make_colormap(seq):
        seq = [(None,) * 3, 0.0] + list(seq) + [1.0, (None,) * 3]
        cdict = {'red': [], 'green': [], 'blue': []}
        for i, item in enumerate(seq):
            if isinstance(item, float):
                r1, g1, b1 = seq[i - 1]
                r2, g2, b2 = seq[i + 1]
                cdict['red'].append([item, r1, r2])
                cdict['green'].append([item, g1, g2])
                cdict['blue'].append([item, b1, b2])
        return mcolors.LinearSegmentedColormap('CustomMap', cdict)

    def SpinD_mesh(xdata, ydata, zdata):
        xi = np.linspace(min(xdata), max(xdata), 200)
        yi = np.linspace(min(ydata), max(ydata), 200)
        xig, yig = np.meshgrid(xi, yi)
        zig = griddata((xdata, ydata), zdata, (xi[None,:], yi[:,None]), method='cubic')
        #zig = griddata((xdata, ydata), zdata, (xig, yig), method='cubic')
        return xig, yig, zig
    #================================================================
    # main
    #================================================================
    fig = plt.figure()
    fig.set_size_inches(6,6)
    plt.subplots_adjust(wspace=0.3,hspace=0.3)
    ax0 = fig.add_subplot(1,1,1)
    ax0.set_xlabel('x / a', fontsize=20)
    ax0.set_ylabel('y / a', fontsize=20)
    ax0.xaxis.set_tick_params(labelsize=15, direction='in', which='both')
    ax0.yaxis.set_tick_params(labelsize=15, direction='in', which='both')
    #ax0.patch.set_facecolor('darkslategrey')
    ax0.patch.set_facecolor((0.165, 0.161, 0.133))
    ax0.patch.set_alpha(0.5)
    #ax0.set_facecolor('tab:olive')
    X,Y = spinlattice.X, spinlattice.Y
    mx, my, mz = spinlattice.mx, spinlattice.my, spinlattice.mz
    # get color for spins depending on s^z
    c =1.0 -(np.arctan2(mz,np.sqrt(1-np.asarray(mz)**2)) + np.pi/2.0)/np.pi
    #c = plt.cm.hsv(c*0.7)
    #cmap = user_defined_colormap()
    cmap = plt.cm.hsv
    # plot results
    #
    #xig, yig, zig = SpinD_mesh(X, Y, c)
    #ax0.imshow(zig, cmap=cmap, extent=[0, spinlattice.size, -spinlattice.size*0.86, spinlattice.size*0.86])

    ax0.scatter(X,Y, marker='.', c=cmap(c*0.7))
    #ax0.quiver(X,Y, mx, my, scale=12, units='width', pivot='middle', width=0.008)
    ax0.quiver(X,Y, mx, my, scale=12, units='width', pivot='middle', color=cmap(c*0.7), width=0.006)
    if details == 'DW':
        DWdetails(ax0)
    elif details == 'Sk':
        Skdetails(ax0)
    elif details == 'SS':
        SSdetails(ax0)
    if zoom != None:
        # BSP for zoom in center : zoom = [spinlattice.size/2, 0, 10]
        xc, yc, dist = zoom
        ax0.set_xlim(xc -dist, xc +dist)
        ax0.set_ylim(yc -dist, yc +dist)
        ax0.set_aspect('equal', 'datalim')
        ax0.margins(0.1)
    else :
        ax0.axis('equal')
    plt.tight_layout()
    cwd = os.getcwd()
    plt.savefig(cwd + '/' + filename)




#=============================================
# function for visualisation of hessian eigenvectors
#=============================================
def show_eigenvector(vectorlattice, zoom=None):
    import matplotlib.pyplot as plt
    from matplotlib.ticker import MultipleLocator, NullFormatter, LogLocator
    # renormalize vectors
    norm = 1./abs(max(vectorlattice.magmom))
    # plot like @show_lattice
    fig = plt.figure()
    fig.set_size_inches(4,4)
    plt.subplots_adjust(left=0.01,right=0.99,top=0.99,bottom=0.01)
    ax0 = fig.add_subplot(1,1,1)
    #ax0.set_xlabel('x / a', fontsize=20)
    #ax0.set_ylabel('y / a', fontsize=20)
    ax0.xaxis.set_tick_params(labelsize=15, direction='in', which='both')
    ax0.yaxis.set_tick_params(labelsize=15, direction='in', which='both')
    ax0.xaxis.set_major_formatter(NullFormatter())
    ax0.yaxis.set_major_formatter(NullFormatter())

    X,Y = vectorlattice.X, vectorlattice.Y
    mx, my, mz = vectorlattice.mx*norm, vectorlattice.my*norm, vectorlattice.mz*norm
    # COLORBAR ACCORDING TO @mz
    #c = 0.5 -(np.arctan2(mz,np.sqrt(1-np.asarray(mz)**2)))
    #c = (c - min(c))/(max(c)-min(c))
    #c = plt.cm.seismic(c)
    # COLORBAR ACCORDING TO INPLANE ANGLE
    c_tmp = []
    for n in range(len(mx)):
        tmp = np.arctan2(my[n], mx[n])
        if tmp < 0. :
            c_tmp.append(tmp +2.*np.pi)
        else :
            c_tmp.append(tmp)
    c_tmp = np.array(c_tmp)
    c = plt.cm.hsv(c_tmp/2./np.pi)
    # OLD COLORBAR VERSION
    #c = np.arctan2(mz,1)
    #if c.ptp() > 0 :
    #    c = (c.ravel() - c.min()) / c.ptp()
    #    c = 1 - np.concatenate((c, np.repeat(c, 2)))
    #    c = plt.cm.seismic(1-c)
    #else :
    #    c = plt.cm.seismic([0 for n in range(len(mz))])
    #ax0.quiver(X,Y, mx, my, scale=15, units='width', pivot='middle', linewidths=1, color=c)
    widths=[1.8 for x in range(mx.size)]
    #ax0.quiver(X,Y, mx, my, scale=14, pivot='middle', linewidths=widths, edgecolors=c, color=c, headwidth=6)
    ax0.quiver(X,Y, mx, my, scale=10, units='width', pivot='middle', color=c, width=0.005)
    #ax0.scatter(X,Y, marker='.', s=4, c=c)
    if zoom != None:
        # BSP for zoom in center : zoom = [spinlattice.size/2, 0, 10]
        xc, yc, dist = zoom
        ax0.set_xlim(xc -dist, xc +dist)
        ax0.set_ylim(yc -dist, yc +dist)
        ax0.set_aspect('equal', 'datalim')
        ax0.margins(0.1)
    else :
        ax0.axis('equal')
    cwd = os.getcwd()
    plt.savefig(cwd + '/vectorlattice.png')


#=============================================
# function for visualisation of hessian eigenvectors
#=============================================
def show_eigenvector_translation(spinlattice, vectorlattice, amplitude):
    import matplotlib.pyplot as plt
    from magnetisations import rotate_to_axis, rotation_matrix, rotate_spins, SpinLattice
    # renormalize vectors
    # plot like @show_lattice

    if spinlattice.size != vectorlattice.size :
        print('original lattice and eigenvector are of different sizes')
        return 0

    fig = plt.figure()
    fig.set_size_inches(8,8)
    plt.subplots_adjust(wspace=0.3,hspace=0.3)
    ax0 = fig.add_subplot(1,1,1)
    ax0.set_xlabel('x / a', fontsize=20)
    ax0.set_ylabel('y / a', fontsize=20)
    ax0.xaxis.set_tick_params(labelsize=15, direction='in', which='both')
    ax0.yaxis.set_tick_params(labelsize=15, direction='in', which='both')
    X,Y = vectorlattice.X, vectorlattice.Y
    mx, my, mz = [], [], []
    # normalise the eigenvectors, so that the bigest corresponds to 1
    norm = 1/max(vectorlattice.magmom)

    # for each spin:
    #   find the axis of rotation as the cross product of the spin and the corresponding component of the eigenvector
    #   rotate each spin in this direction using @amplitude*magnitude(eig vec) as angle
    for m in range(len(spinlattice.mx)):
        axis, ang = rotate_to_axis([vectorlattice.mx[m], vectorlattice.my[m], vectorlattice.mz[m]], [spinlattice.mx[m], spinlattice.my[m], spinlattice.mz[m]])
        mat = rotation_matrix(axis, amplitude*vectorlattice.magmom[m]*norm)
        tempx, tempy, tempz = rotate_spins([spinlattice.mx[m]], [spinlattice.my[m]], [spinlattice.mz[m]], mat)
        mx.append(tempx)
        my.append(tempy)
        mz.append(tempz)

    c = np.arctan2(mz,1)
    if c.ptp() > 0 :
        c = (c.ravel() - c.min()) / c.ptp()
        c = 1 - np.concatenate((c, np.repeat(c, 2)))
        c = plt.cm.hsv(c*0.7)
    else :
        c = plt.cm.hsv([0 for n in range(len(mz))])
    ax0.quiver(X,Y, mx, my, scale=8, units='width', pivot='middle', color=c)
    ax0.scatter(X,Y, marker='.', c=c)
    ax0.axis('equal')
    ax0.set_xlim(spinlattice.size/2 -10, spinlattice.size/2 +10 )
    ax0.set_ylim(-10, 10 )
    #plt.tight_layout()
    cwd = os.getcwd()
    plt.savefig(cwd + '/translated_lattice_amp_'+ str(amplitude) + '.png')







#=============================================
# function to visualize the topological density
# of every unit cell of the lattice.
#=============================================
def show_topodensity(spinlattice, filename='topolodensity.png', zoom=None):
    import matplotlib.pyplot as plt
    from propertyfunctions import get_spherical_area

    fig = plt.figure()
    fig.set_size_inches(6,6)
    plt.subplots_adjust(wspace=0.3,hspace=0.3)
    ax0 = fig.add_subplot(1,1,1)
    ax0.set_xlabel('x / a', fontsize=20)
    ax0.set_ylabel('y / a', fontsize=20)
    ax0.axis('equal')
    ax0.xaxis.set_tick_params(labelsize=15, direction='in', which='both')
    ax0.yaxis.set_tick_params(labelsize=15, direction='in', which='both')

    X,Y = spinlattice.X, spinlattice.Y
    mx, my, mz = spinlattice.mx, spinlattice.my, spinlattice.mz
    size = int(np.sqrt(len(mx)))
    xs, ys, qs = [], [], []
    Q_neg, Q_pos = 0., 0.
    for n in range(size-1):
        for m in range(size-1):
            index1 = n*size + m
            index2 = (n+1)*(size-1) + (m+1) +n
            index3 = (n+1)*(size-1) + (m+2) +n
            index4 = n*size + (m+1)
            S1 = np.asarray([my[index1], mx[index1], mz[index1]])
            S2 = np.asarray([my[index2], mx[index2], mz[index2]])
            S3 = np.asarray([my[index3], mx[index3], mz[index3]])
            S4 = np.asarray([my[index4], mx[index4], mz[index4]])

            # calculate spherical area of the first triangle in unitcell
            area1 = get_spherical_area(S1, S2, S3)
            qs.append(area1)
            xs.append([X[index1], X[index2], X[index3]])
            ys.append([Y[index1], Y[index2], Y[index3]])
            # calculate 2 triangle in unitcell
            area2 = get_spherical_area(S1, S3, S4)
            qs.append(area2)
            xs.append([X[index1], X[index3], X[index4]])
            ys.append([Y[index1], Y[index3], Y[index4]])

            # CHECK FOR POSITIVE AND NEGATIV PARTIAL CHARGES
            #if area1 < 0. :
            #    Q_neg += area1
            #elif area1 > 0. :
            #    Q_pos += area1
            #if area2 < 0. :
            #    Q_neg += area2
            #elif area2 > 0. :
            #    Q_pos += area2

    #print 'Q_neg = ', Q_neg/4./np.pi
    #print 'Q_pos = ', Q_pos/4./np.pi
    Q = sum(qs)/4./np.pi

    #c = np.arctan2(qs,max(max(qs), abs(min(qs))))
    #c = (c.ravel() - c.min()) / c.ptp()
    #c = np.concatenate((c, np.repeat(c, 2)))
    #c = plt.cm.viridis(c)

    max_value = max([abs(q) for q in qs])
    c = plt.cm.seismic(-np.array(qs)/max_value/2.+0.5)
    # ZOOM FOR FASTER PLOTTING
    if zoom != None:
        # BSP for zoom in center : zoom = [spinlattice.size/2, 0, 10]
        xc, yc, dist = zoom
        ax0.set_xlim(xc -dist, xc +dist)
        ax0.set_ylim(yc -dist, yc +dist)
    for p in range(len(qs)):
        if zoom != None :
            if any([xs[p][n] > xc+dist for n in range(3)]) or any([xs[p][n] < xc-dist for n in range(3)]) or any([ys[p][n] > yc+dist for n in range(3)]) or any([ys[p][n] < yc-dist for n in range(3)]) :
                continue
            else :
                ax0.plot(xs[p], ys[p], marker='o', markersize=3, color='k', ls='')
                ax0.fill(xs[p], ys[p], c=c[p])
        else :
            ax0.plot(xs[p], ys[p], marker='o', markersize=3, color='k', ls='')
            ax0.fill(xs[p], ys[p], c=c[p])
    #fig.colorbar(plt.cm.ScalarMappable(), ax=ax0)
    #ax0.set_xlim(spinlattice.size/2 -10, spinlattice.size/2 +10 )
    #ax0.set_ylim(-10, 10)
    plt.tight_layout()
    cwd = os.getcwd()
    plt.savefig(cwd + '/' + filename)


#============================================
# function that creates a histogram of all angles,
# up to 6 shells of neighbours are considered
#============================================
def show_anglehist(spinlattice):
    import matplotlib.pyplot as plt
    #================================================================
    # the fuction runs through all lattice points and checks all neighbours relative positions
    # to avoid double counting for point n just points m\in{n, size} are considered
    # if position matches a shell distant, add the angle to the prober bin
    #================================================================
    def NNangles_hex(splat):
        size =len(splat.mz)
        #eps = np.finfo(np.float32).eps
        bin1, bin2, bin3, bin4, bin5, bin6 = [], [], [], [], [], []
        for n in range(size):
            Rn = [splat.X[n], splat.Y[n]]
            Mn = [splat.mx[n], splat.my[n], splat.mz[n]]
            for m in range(n+1, size):
                Rm = [splat.X[m], splat.Y[m]]
                Mm = [splat.mx[m], splat.my[m], splat.mz[m]]
                d = round(dist(Rn, Rm), 5)
                if d < 4 :
                    if d == 1.0: bin1.append(angle(Mn,Mm))
                    elif d == 1.73205: bin2.append(angle(Mn,Mm))
                    elif d == 2.0: bin3.append(angle(Mn,Mm))
                    elif d == 2.64575: bin4.append(angle(Mn,Mm))
                    elif d == 3.0: bin5.append(angle(Mn,Mm))
                    elif d == 3.46410: bin6.append(angle(Mn,Mm))
        return [np.asarray(bin1), np.asarray(bin2), np.asarray(bin3), np.asarray(bin4), np.asarray(bin5), np.asarray(bin6)]
    #================================================================
    # funtion for the distance of 2 n-dimensional vectos @x, @y
    #================================================================
    def dist(x,y):
        assert len(x) == len(y)
        d = 0.0
        for n in range(len(x)):
            d += (x[n] - y[n])**2
        return np.sqrt(d)
    #================================================================
    # function returns the angle between to n-dimensional vectors in degree
    #================================================================
    def angle(x,y):
        assert len(x) == len(y)
        ang = np.dot(x,y)
        if ang > 1.0 :
            ang = ang - np.finfo(np.float32).eps
        return np.arccos(ang) *180.0/np.pi
    #================================================================
    # main
    #================================================================
    fig=plt.figure()
    fig.set_size_inches(8,8)
    ax0 = fig.add_subplot(1,1,1)
    ax0.set_ylabel('count', fontsize=20)
    ax0.set_xlabel(r'$\phi$ /deg', fontsize=20)
    ax0.xaxis.set_tick_params(labelsize=15, direction='in', which='both')
    ax0.yaxis.set_tick_params(labelsize=15, direction='in', which='both')
    colors = ['red', 'tan', 'lime', 'blue', 'violet', 'darkgreen']
    ks = ['k', 'k', 'k', 'k', 'k', 'k']
    labels = ['nn=1', 'nn=2', 'nn=3', 'nn=4', 'nn=5', 'nn=6']
    bins = NNangles_hex(spinlattice)
    #bins = [np.random.randn(n) for n in [10, 9, 8, 7, 6, 5]]
    ax0.hist(bins, 15, color=colors, density=False, histtype='bar', label=labels)
    ax0.hist(bins, 15, color=ks, density=False, histtype='step', fill=False)
    plt.legend(loc='upper center', fontsize=15)
    plt.tight_layout()




#=====================================================================
# IMPUT: SpinLattice object @sp and a target .png-file with name @image_name
# OUTPUT: povray image of lattice with name  @image_name
# DISCRIBTION: The function uses vapory-library to render a povray image of
#              the given SpinLattice object. The colormap can be choosen to
#              any of plt.cm. Notice that np.arctan2(y,x) has reversed order
#              of entries. According to these anles, every spin is rotated:
#              @theta around y-axis, @phi around z-axis
#=====================================================================
def spinpov(sp, image_name):
    import matplotlib.pyplot as plt
    from magnetisations import SpinLattice
    from matplotlib.colors import LinearSegmentedColormap
    import sys
    sys.path.append('/home/goerzen/libraries/vapory')
    from vapory import Camera, Background, LightSource, Cone, Texture, Pigment, Finish, Scene
    # set standard objects
    # Note that povray uses a left handed coordinate system. Therefore we change
    # it to right handed by defining @up and @right.
    #gratio = 1.61803398875
    gratio = 1.0
    camera = Camera('sky', [0,0,1],
                    'look_at', [sp.size//2, 0,0],
                    'location', [sp.size/2.0,0,sp.size/2.],
                    'up', [0,0,1],
                    'right', [-1,0,0],
                    'angle', 40)
    objects = []
    #objects.append(LightSource([sp.size//2, -sp.size//2, 40], 'color', [1.,1.,1.],  ))
    objects.append(Background( "color", [1,1,1] ))
    objects.append(LightSource([sp.size/2.0, 0, 30], 0.7, 'area_light', (5,0,0), (0,5,0), 5,5, 'adaptive', 1, 'jitter'  ))
    # get the color code according to @sp.mz from plt.cm
    c =1.0 -(np.arctan2(sp.mz,np.sqrt(1-np.asarray(sp.mz)**2)) + np.pi/2.0)/np.pi
    #========================================
    # create your own colorcode, if wanted
    #cmap_name = 'wbr'
    #colors = [(1, 1, 1), (0, 0, 0), (0.545, 0, 0)]
    #cm = LinearSegmentedColormap.from_list(cmap_name, colors, N=265)
    #c = cm(c)
    #========================================
    #c = plt.cm.inferno(c)
    c = plt.cm.hsv(c*0.7)
    # add cons to list of objects. rotation is in [deg]
    for n in range(len(sp.X)):
        theta = np.arccos(sp.mz[n])*180.0/np.pi
        phi = np.arctan2(sp.my[n],sp.mx[n])*180.0/np.pi
        cone = Cone([0.0, 0.0,0.5], 0.0, [0.0, 0.0, -0.5], 0.38,
                    'rotate', [0., theta, 0.],
                    'rotate', [0., 0., phi],
                    'translate', [gratio*sp.X[n], gratio*sp.Y[n], gratio*sp.Z[n]],
                    Texture(Pigment('color', c[n][:-1])),
                    Finish('ambient', 1.0, 'reflection',0.0,'metallic', 0.0))
        objects.append(cone)
    # build the scene and render
    scene = Scene( camera, objects= objects)
    scene.render(image_name, width=700, height=700, antialiasing=0.001)





def show_fft(sp, complex=False):
    from mpl_toolkits import mplot3d
    from mpl_toolkits.mplot3d import proj3d
    import matplotlib.pyplot as plt
    from matplotlib.ticker import MultipleLocator, NullFormatter
    from scipy.interpolate import griddata

    def lattice_fft(sp):
        L = sp.size
        xfft = np.fft.ifft2(np.asarray(sp.mx).reshape(L, L)).transpose()
        xfft = np.fft.fftshift(xfft).reshape(sp.size**2)
        yfft = np.fft.ifft2(np.asarray(sp.my).reshape(L, L)).transpose()
        yfft = np.fft.fftshift(yfft).reshape(sp.size**2)
        zfft = np.fft.ifft2(np.asarray(sp.mz).reshape(L, L)).transpose()
        zfft = np.fft.fftshift(zfft).reshape(sp.size**2)
        return xfft, yfft, zfft

    def SpinD_mesh(xdata, ydata, zdata):
        xi = np.linspace(min(xdata), max(xdata), 200)
        yi = np.linspace(min(ydata), max(ydata), 200)
        xig, yig = np.meshgrid(xi, yi)
        zig = griddata((xdata, ydata), zdata, (xi[None,:], yi[:,None]), method='cubic')
        return xig, yig, zig

    def reciproce_lattice(sp):
        latvec = np.asarray([ [0.5, -0.86602540378, 0.0], [0.5, 0.86602540378, 0.0], [0.0, 0.0, 1.0] ])
        spat = np.dot(latvec[0], np.cross(latvec[1], latvec[2]))
        recvec = np.asarray([ np.cross(latvec[1], latvec[2]), np.cross(latvec[2], latvec[0]), np.cross(latvec[0], latvec[1]) ])*2.0*np.pi/spat
        recX, recY, recZ = [], [], []
        for n in range(sp.size):
            r0 = n*recvec[0]/np.linalg.norm(recvec[0])/sp.size*2
            for m in range(sp.size) :
                r = r0 + m*recvec[1]/np.linalg.norm(recvec[1])/sp.size*2
                recX.append(r[0])
                recY.append(r[1])
                recZ.append(r[2])
        # resize recX, recY to the form of the brillouin zone
        recX = [ (-1.0 +2.0*(x-min(recX))/(max(recX)-min(recX))) for x in recX ]
        recY = [ np.cos(np.pi/6.0)*0.666*(-1.0 +2.0*(y-min(recY))/(max(recY)-min(recY))) for y in recY ]
        return recX, recY, recZ

    def simple_lattice(sp):
        simX, simY, simZ = [], [], []
        for n in range(sp.size):
            r0 = [n,0,0]
            for m in range(sp.size) :
                r = r0 + [0,m,0]
                simX.append(r[0]/sp.size)
                simY.append(r[1]/sp.size)
                simZ.append(r[2]/sp.size)
        return simX, simY, simZ

    fig=plt.figure()
    fig.set_size_inches(8,2)
    ax0 = fig.add_subplot(1,3,1)
    ax1 = fig.add_subplot(1,3,2)
    ax2 = fig.add_subplot(1,3,3)
    plt.subplots_adjust(wspace=0.05,hspace=0.05)
    ax0.set_title(r'FFT(s$_{x}$)')
    ax1.set_title(r'FFT(s$_{y}$)')
    ax2.set_title(r'FFT(s$_{z}$)')
    ax0.axis('equal')
    ax1.axis('equal')
    ax2.axis('equal')
    ax0._axis3don = False
    ax1._axis3don = False
    ax2._axis3don = False
    ax0.xaxis.set_tick_params(labelsize=8, direction='in', which='both')
    ax0.yaxis.set_tick_params(labelsize=8, direction='in', which='both')
    ax1.xaxis.set_tick_params(labelsize=8, direction='in', which='both')
    ax1.yaxis.set_tick_params(labelsize=8, direction='in', which='both')
    ax2.xaxis.set_tick_params(labelsize=8, direction='in', which='both')
    ax2.yaxis.set_tick_params(labelsize=8, direction='in', which='both')

    ax0.spines['left'].set_position('center')
    ax0.spines['bottom'].set_position('center')
    # Eliminate upper and right axes
    ax0.spines['right'].set_color('none')
    ax0.spines['top'].set_color('none')
    # Show ticks in the left and lower axes only
    ax0.xaxis.set_ticks_position('bottom')
    ax0.yaxis.set_ticks_position('left')

    ax1.spines['left'].set_position('center')
    ax1.spines['bottom'].set_position('center')
    # Eliminate upper and right axes
    ax1.spines['right'].set_color('none')
    ax1.spines['top'].set_color('none')
    # Show ticks in the left and lower axes only
    ax1.xaxis.set_ticks_position('bottom')
    ax1.yaxis.set_ticks_position('left')

    ax2.spines['left'].set_position('center')
    ax2.spines['bottom'].set_position('center')
    # Eliminate upper and right axes
    ax2.spines['right'].set_color('none')
    ax2.spines['top'].set_color('none')
    # Show ticks in the left and lower axes only
    ax2.xaxis.set_ticks_position('bottom')
    ax2.yaxis.set_ticks_position('left')

    ax0.set_xticks( np.array([-1, -0.666, 0.666, 1]))
    ax0.set_xticklabels(['-M', '-K', 'K', 'M'])
    ax0.yaxis.set_major_formatter(NullFormatter())

    ax1.set_xticks( np.array([-1, -0.666, 0.666, 1]))
    ax1.set_xticklabels(['-M', '-K', 'K', 'M'])
    ax1.yaxis.set_major_formatter(NullFormatter())

    ax2.set_xticks( np.array([-1, -0.666, 0.666, 1]))
    ax2.set_xticklabels(['-M', '-K', 'K', 'M'])
    ax2.yaxis.set_major_formatter(NullFormatter())

    # main call
    xfft, yfft, zfft = lattice_fft(sp)
    # get the reciproke coordinates
    recX, recY, recZ = reciproce_lattice(sp)
    #recX, recY, recZ = simple_lattice(sp)
    xigX, yigX, zigX = SpinD_mesh(recX, recY, np.absolute(xfft))
    xigY, yigY, zigY = SpinD_mesh(recX, recY, np.absolute(yfft))
    xigZ, yigZ, zigZ = SpinD_mesh(recX, recY, np.absolute(zfft))
    lim = max(max(np.absolute(sp.mx)), max(np.absolute(sp.my)), max(np.absolute(sp.mz)))
    # note: if colors shall be interpolated use @shading='gouraud'
    ax0.pcolormesh(xigX, yigX, zigX, shading='gouraud', cmap='coolwarm')
    ax1.pcolormesh(xigY, yigY, zigY, shading='gouraud', cmap='coolwarm')
    ax2.pcolormesh(xigZ, yigZ, zigZ, shading='gouraud', cmap='coolwarm')

    ax0.text(0.55, np.cos(np.pi/6.0)*0.666*0.5, "M", ha="center", va="center", rotation=0, size=8)
    ax0.plot(0.5, np.cos(np.pi/6.0)*0.666*0.5, marker='o', color='k', markersize='3')

    ax1.text(0.55, np.cos(np.pi/6.0)*0.666*0.5, "M", ha="center", va="center", rotation=0, size=8)
    ax1.plot(0.5, np.cos(np.pi/6.0)*0.666*0.5, marker='o', color='k', markersize='3')

    ax2.text(0.55, np.cos(np.pi/6.0)*0.666*0.5, "M", ha="center", va="center", rotation=0, size=8)
    ax2.plot(0.5, np.cos(np.pi/6.0)*0.666*0.5, marker='o', color='k', markersize='3')

    plt.tight_layout()
    cwd = os.getcwd()
    plt.savefig(cwd + '/spinlattice_fft.png')

    # make a new plot with real and imaginary components
    if complex :

        fig2=plt.figure()
        fig2.set_size_inches(8,2)
        ax0 = fig2.add_subplot(2,3,1)
        ax1 = fig2.add_subplot(2,3,2)
        ax2 = fig2.add_subplot(2,3,3)
        ax01 = fig2.add_subplot(2,3,4)
        ax11 = fig2.add_subplot(2,3,5)
        ax21 = fig2.add_subplot(2,3,6)
        plt.subplots_adjust(wspace=0.05,hspace=0.05)
        ax0.set_title(r'$\Re$(FFT(s$_{x}$))')
        ax1.set_title(r'$\Re$(FFT(s$_{y}$))')
        ax2.set_title(r'$\Re$(FFT(s$_{z}$))')
        ax01.set_title(r'$\Im$(FFT(s$_{x}$))')
        ax11.set_title(r'$\Im$(FFT(s$_{y}$))')
        ax21.set_title(r'$\Im$(FFT(s$_{z}$))')
        ax0.axis('equal')
        ax1.axis('equal')
        ax2.axis('equal')
        ax01.axis('equal')
        ax11.axis('equal')
        ax21.axis('equal')
        ax0._axis3don = False
        ax1._axis3don = False
        ax2._axis3don = False
        ax01._axis3don = False
        ax11._axis3don = False
        ax21._axis3don = False
        ax0.xaxis.set_tick_params(labelsize=15, direction='in', which='both')
        ax0.yaxis.set_tick_params(labelsize=15, direction='in', which='both')
        ax1.xaxis.set_tick_params(labelsize=15, direction='in', which='both')
        ax1.yaxis.set_tick_params(labelsize=15, direction='in', which='both')
        ax2.xaxis.set_tick_params(labelsize=15, direction='in', which='both')
        ax2.yaxis.set_tick_params(labelsize=15, direction='in', which='both')
        ax01.xaxis.set_tick_params(labelsize=15, direction='in', which='both')
        ax01.yaxis.set_tick_params(labelsize=15, direction='in', which='both')
        ax11.xaxis.set_tick_params(labelsize=15, direction='in', which='both')
        ax11.yaxis.set_tick_params(labelsize=15, direction='in', which='both')
        ax21.xaxis.set_tick_params(labelsize=15, direction='in', which='both')
        ax21.yaxis.set_tick_params(labelsize=15, direction='in', which='both')

        ax0.spines['left'].set_position('center')
        ax0.spines['bottom'].set_position('center')
        # Eliminate upper and right axes
        ax0.spines['right'].set_color('none')
        ax0.spines['top'].set_color('none')
        # Show ticks in the left and lower axes only
        ax0.xaxis.set_ticks_position('bottom')
        ax0.yaxis.set_ticks_position('left')

        ax1.spines['left'].set_position('center')
        ax1.spines['bottom'].set_position('center')
        # Eliminate upper and right axes
        ax1.spines['right'].set_color('none')
        ax1.spines['top'].set_color('none')
        # Show ticks in the left and lower axes only
        ax1.xaxis.set_ticks_position('bottom')
        ax1.yaxis.set_ticks_position('left')

        ax2.spines['left'].set_position('center')
        ax2.spines['bottom'].set_position('center')
        # Eliminate upper and right axes
        ax2.spines['right'].set_color('none')
        ax2.spines['top'].set_color('none')
        # Show ticks in the left and lower axes only
        ax2.xaxis.set_ticks_position('bottom')
        ax2.yaxis.set_ticks_position('left')

        ax01.spines['left'].set_position('center')
        ax01.spines['bottom'].set_position('center')
        # Eliminate upper and right axes
        ax01.spines['right'].set_color('none')
        ax01.spines['top'].set_color('none')
        # Show ticks in the left and lower axes only
        ax01.xaxis.set_ticks_position('bottom')
        ax01.yaxis.set_ticks_position('left')

        ax11.spines['left'].set_position('center')
        ax11.spines['bottom'].set_position('center')
        # Eliminate upper and right axes
        ax11.spines['right'].set_color('none')
        ax11.spines['top'].set_color('none')
        # Show ticks in the left and lower axes only
        ax11.xaxis.set_ticks_position('bottom')
        ax11.yaxis.set_ticks_position('left')

        ax21.spines['left'].set_position('center')
        ax21.spines['bottom'].set_position('center')
        # Eliminate upper and right axes
        ax21.spines['right'].set_color('none')
        ax21.spines['top'].set_color('none')
        # Show ticks in the left and lower axes only
        ax21.xaxis.set_ticks_position('bottom')
        ax21.yaxis.set_ticks_position('left')

        ax0.set_xticks( np.array([-1, -0.666, 0.666, 1]))
        ax0.set_xticklabels(['-M', '-K', 'K', 'M'])
        ax0.yaxis.set_major_formatter(NullFormatter())

        ax1.set_xticks( np.array([-1, -0.666, 0.666, 1]))
        ax1.set_xticklabels(['-M', '-K', 'K', 'M'])
        ax1.yaxis.set_major_formatter(NullFormatter())

        ax2.set_xticks( np.array([-1, -0.666, 0.666, 1]))
        ax2.set_xticklabels(['-M', '-K', 'K', 'M'])
        ax2.yaxis.set_major_formatter(NullFormatter())

        ax01.set_xticks( np.array([-1, -0.666, 0.666, 1]))
        ax01.set_xticklabels(['-M', '-K', 'K', 'M'])
        ax01.yaxis.set_major_formatter(NullFormatter())

        ax11.set_xticks( np.array([-1, -0.666, 0.666, 1]))
        ax11.set_xticklabels(['-M', '-K', 'K', 'M'])
        ax11.yaxis.set_major_formatter(NullFormatter())

        ax21.set_xticks( np.array([-1, -0.666, 0.666, 1]))
        ax21.set_xticklabels(['-M', '-K', 'K', 'M'])
        ax21.yaxis.set_major_formatter(NullFormatter())

        xigX_re, yigX_re, zigX_re = SpinD_mesh(recX, recY, [z.real for z in xfft])
        xigY_re, yigY_re, zigY_re = SpinD_mesh(recX, recY, [z.real for z in yfft])
        xigZ_re, yigZ_re, zigZ_re = SpinD_mesh(recX, recY, [z.real for z in zfft])

        xigX_im, yigX_im, zigX_im = SpinD_mesh(recX, recY, [z.imag for z in xfft])
        xigY_im, yigY_im, zigY_im = SpinD_mesh(recX, recY, [z.imag for z in yfft])
        xigZ_im, yigZ_im, zigZ_im = SpinD_mesh(recX, recY, [z.imag for z in zfft])

        ax0.pcolormesh(xigX_re, yigX_re, zigX_re, shading='gouraud', cmap='seismic')
        ax1.pcolormesh(xigY_re, yigY_re, zigY_re, shading='gouraud', cmap='seismic')
        ax2.pcolormesh(xigZ_re, yigZ_re, zigZ_re, shading='gouraud', cmap='seismic')

        ax01.pcolormesh(xigX_im, yigX_im, zigX_im, shading='gouraud', cmap='seismic')
        ax11.pcolormesh(xigY_im, yigY_im, zigY_im, shading='gouraud', cmap='seismic')
        ax21.pcolormesh(xigZ_im, yigZ_im, zigZ_im, shading='gouraud', cmap='seismic')


    plt.tight_layout()
    #cwd = os.getcwd()
    #plt.savefig(cwd + '/spinlattice_fft.png')
    plt.show()









def show_hilbert_transform(sp):
    from mpl_toolkits import mplot3d
    from mpl_toolkits.mplot3d import proj3d
    import matplotlib.pyplot as plt
    from matplotlib.ticker import MultipleLocator, NullFormatter
    from scipy.interpolate import griddata
    from scipy import signal


    def lattice_fft(sp):
        L = sp.size
        xfft = np.fft.ifft2(np.asarray(sp.mx).reshape(L, L)).transpose()
        xfft = np.fft.fftshift(xfft).reshape(sp.size**2)
        yfft = np.fft.ifft2(np.asarray(sp.my).reshape(L, L)).transpose()
        yfft = np.fft.fftshift(yfft).reshape(sp.size**2)
        zfft = np.fft.ifft2(np.asarray(sp.mz).reshape(L, L)).transpose()
        zfft = np.fft.fftshift(zfft).reshape(sp.size**2)
        return xfft, yfft, zfft

    def lattice_hilbert_trafo(sp):
        L = sp.size

        xfft, yfft, zfft = lattice_fft(sp)

        xfft = signal.hilbert2(np.asarray(np.absolute(xfft)).reshape(L, L))
        xfft = np.imag(xfft).reshape(sp.size**2)
        #xfft = np.fft.fftshift(xfft).reshape(sp.size**2)
        #xfft = np.fft.fftshift(xfft).reshape(sp.size**2)
        yfft = signal.hilbert2(np.asarray(np.absolute(yfft)).reshape(L, L))
        yfft = np.imag(yfft).reshape(sp.size**2)
        #yfft = np.fft.fftshift(yfft).reshape(sp.size**2)
        zfft = signal.hilbert2(np.asarray(np.absolute(zfft)).reshape(L, L))
        zfft = np.imag(zfft).reshape(sp.size**2)
        #zfft = np.fft.fftshift(zfft).reshape(sp.size**2)
        return xfft, yfft, zfft

    def SpinD_mesh(xdata, ydata, zdata):
        xi = np.linspace(min(xdata), max(xdata), 200)
        yi = np.linspace(min(ydata), max(ydata), 200)
        #xi = np.linspace(-1.17, 1.17, 200)
        #yi = np.linspace(-np.cos(np.pi/6.0), np.cos(np.pi/6.0), 200)
        xig, yig = np.meshgrid(xi, yi)
        zig = griddata((xdata, ydata), zdata, (xi[None,:], yi[:,None]), method='cubic')
        return xig, yig, zig

    def reciproce_lattice(sp):
        latvec = np.asarray([ [0.5, -0.86602540378, 0.0], [0.5, 0.86602540378, 0.0], [0.0, 0.0, 1.0] ])
        spat = np.dot(latvec[0], np.cross(latvec[1], latvec[2]))
        recvec = np.asarray([ np.cross(latvec[1], latvec[2]), np.cross(latvec[2], latvec[0]), np.cross(latvec[0], latvec[1]) ])*2.0*np.pi/spat
        recX, recY, recZ = [], [], []
        for n in range(sp.size):
            r0 = n*recvec[0]/np.linalg.norm(recvec[0])/sp.size*2
            for m in range(sp.size) :
                r = r0 + m*recvec[1]/np.linalg.norm(recvec[1])/sp.size*2
                recX.append(r[0])
                recY.append(r[1])
                recZ.append(r[2])
        # resize recX, recY to the form of the brillouin zone
        recX = [ (-1.0 +2.0*(x-min(recX))/(max(recX)-min(recX))) for x in recX ]
        recY = [ np.cos(np.pi/6.0)*0.666*(-1.0 +2.0*(y-min(recY))/(max(recY)-min(recY))) for y in recY ]
        return recX, recY, recZ

    # main call
    xfft, yfft, zfft = lattice_hilbert_trafo(sp)

    fig=plt.figure()
    fig.set_size_inches(8,4)
    ax0 = fig.add_subplot(1,3,1)
    ax1 = fig.add_subplot(1,3,2)
    ax2 = fig.add_subplot(1,3,3)
    plt.subplots_adjust(wspace=0.05,hspace=0.05)
    ax0.set_title(r'FFT(s$_{x}$)')
    ax1.set_title(r'FFT(s$_{y}$)')
    ax2.set_title(r'FFT(s$_{z}$)')
    ax0.axis('equal')
    ax1.axis('equal')
    ax2.axis('equal')
    ax0._axis3don = False
    ax1._axis3don = False
    ax2._axis3don = False
    ax0.xaxis.set_tick_params(labelsize=15, direction='in', which='both')
    ax0.yaxis.set_tick_params(labelsize=15, direction='in', which='both')
    ax1.xaxis.set_tick_params(labelsize=15, direction='in', which='both')
    ax1.yaxis.set_tick_params(labelsize=15, direction='in', which='both')
    ax2.xaxis.set_tick_params(labelsize=15, direction='in', which='both')
    ax2.yaxis.set_tick_params(labelsize=15, direction='in', which='both')

    ax0.spines['left'].set_position('center')
    ax0.spines['bottom'].set_position('center')
    # Eliminate upper and right axes
    ax0.spines['right'].set_color('none')
    ax0.spines['top'].set_color('none')
    # Show ticks in the left and lower axes only
    ax0.xaxis.set_ticks_position('bottom')
    ax0.yaxis.set_ticks_position('left')

    ax1.spines['left'].set_position('center')
    ax1.spines['bottom'].set_position('center')
    # Eliminate upper and right axes
    ax1.spines['right'].set_color('none')
    ax1.spines['top'].set_color('none')
    # Show ticks in the left and lower axes only
    ax1.xaxis.set_ticks_position('bottom')
    ax1.yaxis.set_ticks_position('left')

    ax2.spines['left'].set_position('center')
    ax2.spines['bottom'].set_position('center')
    # Eliminate upper and right axes
    ax2.spines['right'].set_color('none')
    ax2.spines['top'].set_color('none')
    # Show ticks in the left and lower axes only
    ax2.xaxis.set_ticks_position('bottom')
    ax2.yaxis.set_ticks_position('left')

    ax0.set_xticks( np.array([-1, -0.666, 0.666, 1]))
    ax0.set_xticklabels(['-M', '-K', 'K', 'M'])
    ax0.yaxis.set_major_formatter(NullFormatter())

    ax1.set_xticks( np.array([-1, -0.666, 0.666, 1]))
    ax1.set_xticklabels(['-M', '-K', 'K', 'M'])
    ax1.yaxis.set_major_formatter(NullFormatter())

    ax2.set_xticks( np.array([-1, -0.666, 0.666, 1]))
    ax2.set_xticklabels(['-M', '-K', 'K', 'M'])
    ax2.yaxis.set_major_formatter(NullFormatter())

    # get the reciproke coordinates
    recX, recY, recZ = reciproce_lattice(sp)
    #recX, recY, recZ = simple_lattice(sp)
    xigX, yigX, zigX = SpinD_mesh(recX, recY, xfft)
    xigY, yigY, zigY = SpinD_mesh(recX, recY, yfft)
    xigZ, yigZ, zigZ = SpinD_mesh(recX, recY, zfft)
    lim = max(max(np.absolute(sp.mx)), max(np.absolute(sp.my)), max(np.absolute(sp.mz)))
    # note: if colors shall be interpolated use @shading='gouraud'
    ax0.pcolormesh(xigX, yigX, zigX, shading='gouraud', cmap='coolwarm')
    ax1.pcolormesh(xigY, yigY, zigY, shading='gouraud', cmap='coolwarm')
    ax2.pcolormesh(xigZ, yigZ, zigZ, shading='gouraud', cmap='coolwarm')

    ax0.text(0.55, np.cos(np.pi/6.0)*0.666*0.5, "M", ha="center", va="center", rotation=0, size=15)
    ax0.plot(0.5, np.cos(np.pi/6.0)*0.666*0.5, marker='o', color='k', markersize='3')

    ax1.text(0.55, np.cos(np.pi/6.0)*0.666*0.5, "M", ha="center", va="center", rotation=0, size=15)
    ax1.plot(0.5, np.cos(np.pi/6.0)*0.666*0.5, marker='o', color='k', markersize='3')

    ax2.text(0.55, np.cos(np.pi/6.0)*0.666*0.5, "M", ha="center", va="center", rotation=0, size=15)
    ax2.plot(0.5, np.cos(np.pi/6.0)*0.666*0.5, marker='o', color='k', markersize='3')

    plt.tight_layout()
    plt.show()



# -*- coding: utf-8 -*-
r"""
abstract base class for analyzing the test results
"""
from abc import ABC, abstractmethod


class IAnalyzer(ABC):
    r"""
    Abstract base class for evaluating and analyzing test results.
    """
    @abstractmethod
    def __call__(self, *args, **kwargs) -> bool:
        r"""
        Analyzes the test

        Returns:
            A boolean whether this test failed
        """


# -*- coding: utf-8 -*-
r"""
module with functions:
- write eigenvector
"""
from pathlib import Path
from typing import List, TypeVar
import numpy as np

PathLike = TypeVar("PathLike", str, Path)


def write_vector(vec: List[np.ndarray], outfile: PathLike) -> None:
    r"""
    Writes eigenvector to file

    Args:
        vec(List[np.ndarray]): eigenvector in format [x, y, z, vx, vy, vz, |v|]
        outfile(PathLike): name of the created out file
    """
    wfile = open(outfile, 'a')
    for (idx, line) in enumerate(vec[0]):
        wfile.write(str(vec[0][idx]) + ' ' + str(vec[1][idx]) + ' ' + str(vec[2][idx]) + ' ' + str(vec[3][idx]) +
                    ' ' + str(vec[4][idx]) + ' ' + str(vec[5][idx]) + ' ' + str(vec[6][idx]) + '\n')
    wfile.close()



import numpy as np
from python3.magnetisations import SpinLattice, MultiLayer
from python3.visualizations import *
from python3.shell_commands import *
from python3.spinD_kiel_jobs import *





def run_example_multilayer_GNEB():
    # SET UP MULTILAYER WITH 2 LAYERS CONTAINING SKYRMIONS AS INIT STATE
    c, w = 2., 3.
    helicity, vorticity = 0., 1.
    size, key, magmom1 = 50, 'hex', 1.
    # center is the same for both Skyrmions
    center = [size//2, 0]
    a1_0=2.0/3.0
    a2_0=1.0/3.0
    magmom2= 2.0
    SL0_init = SpinLattice(size=size, key=key, path=None, magmom=magmom1)
    SL0_init.addSkyrmion(center, vorticity, helicity, c, w)
    SL1_init = SpinLattice(size=size, key=key, path=None, magmom=magmom2, a1_0=a1_0, a2_0=a2_0)
    SL1_init.addSkyrmion(center, vorticity, helicity, c, w)
    ML_init =  MultiLayer(number_layers=2)
    ML_init.add_layer(SL0_init, index_layer=0, zpos_uc=0.)
    ML_init.add_layer(SL1_init, index_layer=1, zpos_uc=3.8302600704)

    # SET UP MULTILAYER WITH 2 FM LAYERS AS FINAL STATE
    SL0_final = SpinLattice(size=size, key=key, path=None, magmom=magmom1)
    SL1_final = SpinLattice(size=size, key=key, path=None, magmom=magmom2, a1_0=a1_0, a2_0=a2_0)
    ML_final =  MultiLayer(number_layers=2)
    ML_final.add_layer(SL0_final, index_layer=0, zpos_uc=0.)
    ML_final.add_layer(SL1_final, index_layer=1, zpos_uc=3.8302600704)

    # CREATE DIRECTORY FOR TEST CALCULATION AND DEFINE DIRECTORY CONTAINING INPUTS
    images = 20
    iterations = 1
    input_directory = 'input_data'
    calc_directory = 'calculation_multilayer_GNEB'
    make_directory(calc_directory)

    # START CALCULATION FOR MAGNETIC FIELD B=0
    Bfield = 0.
    adjust_parameter("H_ext", "0.0  0.0  "+str(Bfield), input_directory + "/inp")
    start_GNEB(ML_init, ML_final, calc_directory, input_directory, images, iterations, use_path=None)



# -*- coding: utf-8 -*-
r"""
module for finding local extrema in data. Useful for example in:

- vector translation / mode following: The vector translation produces an outputfile vt_energy.dat. Sometimes one is
interested in investigating the local minima and maxima of the behaviour along the mode in more detail. Therefore one
wants to know the step at which the extremum occurs.
"""
from pathlib import Path
from typing import List, TypeVar, Union, Tuple
import numpy as np
import pandas as pd
from scipy.signal import argrelextrema

PathLike = TypeVar("PathLike", str, Path)


class CFindLocalExtremaModeFollowing:
    r"""
    Class for finding local extrema in vt_energy.dat
    """
    def __init__(self, data: pd.DataFrame = None, filename: PathLike = 'vt_energy.dat') -> None:
        r"""
        Initializes the localization of local minima and maxima within vt_energy.dat

        Args:
            data(DataFrame): data frame containing columns steps and energy. If None the file vt_energy.in is tried to
            read. Another filename can be provided also.
        """
        if data is None:
            self._df = pd.read_csv(filename, usecols=[0, 1], sep=r"\s+", names=['step', 'energy'], index_col=False)
        else:
            self._df = data

        self._energy = self._df['energy'].to_numpy()
        self._step = self._df['step'].to_numpy()

    def __call__(self, method: str = 'scipy', writeoutfile: bool = False) -> Tuple[np.ndarray, np.ndarray]:
        r"""
        Calls analyzing the energy whether local minima exists.

        Args:
            method(str): method for finding the relative extrema. Default is the use of scipys argrelextrema.
            writeoutfile(bool): Whether results shall be written to file.

        Returns:
            array of indices of local maxima and array of indices of local minima
        """
        if method == 'scipy':
            maxs = argrelextrema(self._energy, np.greater)[0]
            mins = argrelextrema(self._energy, np.less)[0]
            if writeoutfile:
                with open('vt_extrema.dat', 'w') as f:
                    f.write('step energy type \n')
                    for mx in maxs:
                        f.write(f'{self._step[mx]} {self._energy[mx]} max \n')
                    for mn in mins:
                        f.write(f'{self._step[mn]} {self._energy[mn]} min \n')
            return maxs, mins
        else:
            raise NotImplementedError('other methods not yet available!')



from examples.functions_analytic_examples import *



#=======================================================
# REMOVE AND ADD COMMENTS TO RUN EXAMPLES
#=======================================================
def main():

    #run_example_charge_and_chargedensity()

    #run_example_setup_cycloidal_1qstates()

    #run_example_setup_domainwall_get_width()

    #run_example_setup_double_domainwall()

    #run_example_setup_skyrmion_get_radius()

    #run_example_setup_multilayer_hex()

    #run_example_setup_multilayer_quad()

    #run_example_setup_bimeron()

    #run_example_setup_nanoskyrmionlattice()

    #run_example_uudd_by_superposition()

    #run_example_setup_AFM_skyrmion()

    run_example_setup_chimera_state()


if __name__ == '__main__' :
     main()
     display_plots()





import sys
from scipy import constants
from scipy.optimize import curve_fit
import matplotlib.pyplot as plt
import numpy as np

from python3.shell_commands import *


#======================================================================
# configure dyna.in
dt = 1.0
dur = 50000
damping = 0.
integrator = 6

# configure inp
H_z = 1.0
gra_log_freq = 500

# configure lattice.in
number_spin = 1
magn_mom = 1.0




class MagnConstants:
    #==================================================================
    #==================================================================

    def __init__(self):
        # for the gyramagnetic ratio
        self.gJ = 2.0
        self.gyro_ratio = self.gJ* constants.e/constants.m_e/2.

        # for the bohr magneton
        self.mu_B = 5.7883818012e-05






class FileConfiguration:
    #==================================================================
    #==================================================================

    def config_inp(self, H_z, gra_log_freq):
        #==============================================================
        adjust_parameter("H_ext", "0.0  0.0  "+str(H_z),"inp")
        adjust_parameter("Gra_log", ' .T. ' +str(gra_log_freq), "inp")


    def config_dyna(self, dt, dur, damp, integrator):
        #==============================================================
        adjust_parameter("timestep", str(dt), "dyna.in")
        adjust_parameter("duration", str(dur), "dyna.in")
        adjust_parameter("damping", str(damp), "dyna.in")
        adjust_parameter("integration", str(integrator), "dyna.in")


    def config_lattice(self, num_spin, magn_mom):
        #==============================================================
        # creates a quadratic lattice with size (num_spin)x(num_spin)
        adjust_parameter("Nsize", " " +str(number_spin) + " " + str(number_spin) + " 1", "lattice.in")


    def config_spiral_start(self):
        #==============================================================
        adjust_parameter("qvec", "0.0  0.0  0.0", "spiral.start")




class FitFunctions():
    #==================================================================
    #==================================================================

    def __init__(self, H_z):
        #==================================================================
        self.H_z = H_z
        # gyromagnetic ratio, with g=1
        self.gyro_ratio_alternative = 1.760859e11
        self.gyro_ratio = constants.e/constants.m_e

    def sin_fit(self, x,f):
        #==================================================================
        return np.sin(f*x)

    def cos_fit(self, x,f):
        #==================================================================
        return np.cos(f*x)

    def tanh_fit(self, x, lamb):
        #==================================================================
        return np.tanh(self.gyro_ratio* self.H_z* x* lamb)






def plot_precission(H_z, damping, dt):
    #==================================================================
    #create figure-object
    fig=plt.figure()
    fig.set_size_inches(6,5)
    plt.subplots_adjust(wspace=0.35,hspace=0.25)
    #create subplot
    ax0=fig.add_subplot(1,1,1)
    # titles
    #labels
    ax0.set_ylabel(r'abs. moment / $\mu_B$')
    ax0.set_xlabel('time /s')

    # get number of files
    number_SpinSTM_files = int(dur/gra_log_freq)-1
    # arrays for coordinats
    x, y, z, T = [], [], [], []

    for i in range(number_SpinSTM_files-1):
        data = np.loadtxt('SpinSTM_' + str(i) + '.dat')
        x.append(data[3])
        y.append(data[4])
        z.append(data[5])
        #!!!!!ATTENTION!!!!!!!!
        # T is scaled by 0.5, so 2 timesteps are 1 timestep on correct time axes
        T.append(dt*i*gra_log_freq*0.5e-15*2)

    # initiate FitFunctions
    fits = FitFunctions(H_z)
    magn = MagnConstants()

    # fit z-data, x-data, y-data, depending on damping is zero or not
    if damping != 0.0 :
        popt0,pcov0 = curve_fit(fits.tanh_fit, T, z, p0=[damping], maxfev=5000)
        ax0.plot(T, fits.tanh_fit(np.asarray(T), *popt0), 'k-')
        print('The damping is : damp=', popt0[0])
        print('And should be  : damp=', damping)
        print('relative deviation |(d_c-d)/d_c|=', abs((popt0[0]-damping)/damping))
    else :
        popt1,pcov1 = curve_fit(fits.cos_fit, T, x, p0=[magn.gyro_ratio*H_z], maxfev=5000)
        popt2,pcov2 = curve_fit(fits.sin_fit, T, y, p0=[magn.gyro_ratio*H_z], maxfev=5000)
        ax0.plot(T, fits.cos_fit(np.asarray(T), *popt1), 'g-')
        ax0.plot(T, fits.sin_fit(np.asarray(T), *popt2), 'r-')
        print('We got Lamorfrequency of : w_c=', popt1[0])
        print('And it should be :         w_c=', magn.gyro_ratio*H_z)
        print('relative deviation  |(w_c-w)/w_c|=', abs((magn.gyro_ratio*H_z - popt1[0])/magn.gyro_ratio*H_z))

    # plot data
    ax0.plot(T, z, 'b+', label=r'S$_z$')
    ax0.plot(T, x, 'g+', label=r'S$_x$')
    ax0.plot(T, y, 'r+', label=r'S$_y$')

    # finishing
    plt.legend(loc='upper right', fontsize=10)
    plt.savefig('plot_test_precission.pdf')
    plt.show()







def main(H_z, gra_log_freq, dt, dur, damp, num_spin, magn_mom, integrator):
    #==================================================================
    make_directory('test_calculation')
    copy_element('input_data/inp', 'test_calculation')
    copy_element('input_data/dyna.in', 'test_calculation')
    copy_element('input_data/lattice.in', 'test_calculation')
    copy_element('input_data/spiral.start', 'test_calculation')
    copy_element('input_data/simu.in', 'test_calculation')

    with change_directory('test_calculation'):
        # load class FileConfiguration
        config = FileConfiguration()
        # configure inputfiles in actuell calculation folder
        config.config_inp(H_z, gra_log_freq)
        config.config_dyna(dt, dur, damp, integrator)
        config.config_lattice(num_spin, magn_mom)
        config.config_spiral_start()

        # execute calculation and wait for it to finish
        call_spin()
        while not os.path.isfile("SpinSTM_end.dat") :
            time.sleep(5)

        # present results
        plot_precission(H_z, damping, dt)





if __name__ == '__main__':
    main(H_z, gra_log_freq, dt, dur, damping, number_spin, magn_mom, integrator)



import numpy as np

#============================================
# function for the topological charge of a lattice, see B.Berg and M.Luescher
# works just on a lattice constructed in the way
# the class 'Lattices' does. For other constructions
# S1,S2,S3,S4 have to be assigned disfferently.
# But the calculation of the spherical area is nice in every way.
# Input : mag = [my, mx,mz]
# Output : float Q
#============================================
def topocharge(mag):
    Q = 0
    size = int(np.sqrt(len(mag[0])))
    for n in range(size-1):
        for m in range(size-1):
            index1 = n*size + m
            index2 = (n+1)*(size-1) + (m+1) +n
            index3 = (n+1)*(size-1) + (m+2) +n
            index4 = n*size + (m+1)
            S1 = np.asarray([mag[0][index1], mag[1][index1], mag[2][index1]])
            S2 = np.asarray([mag[0][index2], mag[1][index2], mag[2][index2]])
            S3 = np.asarray([mag[0][index3], mag[1][index3], mag[2][index3]])
            S4 = np.asarray([mag[0][index4], mag[1][index4], mag[2][index4]])
            area1 = get_spherical_area(S1, S2, S3)
            area2 = get_spherical_area(S1, S3, S4)
            Q += area1 + area2
    return Q/4.0/np.pi
#============================================
# function returns the spherical area created by three spins
# the sign of the area is the sign of the spat product S1*(S2xS3)
# inspired by implementation of B. Dupe
#============================================
def get_spherical_area(S1, S2, S3):
    con = 1 + np.dot(S1,S2) + np.dot(S2,S3) + np.dot(S3,S1)
    # FM CASE
    if con >= 4.-10e-8 :
        area = 0.
    else :
        mat = [S1, S2, S3]
        det_mat = np.linalg.det(mat)
        # EXCEPTIONAL CASE
        if abs(con) <= 10e-8 and abs(det_mat) <=10e-8:
            area = 0.
        # NON-EXCEPTIONAL NON-COLINEAR CASE
        else :
            area = 2.0*np.arctan2(det_mat, con)
    return area


#============================================
# funktion for calculation of the skyrmion radius.
# method adopted from S.v. Malottki
# Input: SpinLattice object
#============================================
def skyrmionradius(splat):
    from scipy.optimize import curve_fit, fmin
    #============================================
    # 2d fit function for the skyrmion profile
    #============================================
    def sk_2dprofile(R, r0x, r0y, c, w):
        r = np.sqrt((R[0]-r0x)**2 + (R[1]-r0y)**2)
        mzs = np.cos(theta(r,c,w))
        return mzs
    #============================================
    # analytic function for the azimuthal angle @theta of a skyrmion
    # and its derivation @dtheta_dr
    # @r is the distance from the centrum
    # @c, @w the fit parameters
    #============================================
    def theta(r,c,w):
        comp1 = np.arcsin(np.tanh((-abs(r) -c)*2.0/w))
        comp2 = np.arcsin(np.tanh((-abs(r) +c)*2.0/w))
        return np.pi +comp1 +comp2
    def dtheta_dr(r,c,w):
        comp1 = 2.0*np.sqrt(-np.tanh(2.0*(-c + abs(r))/w)**2 + 1)/w
        comp2 = 2.0*np.sqrt(-np.tanh(2.0*( c + abs(r))/w)**2 + 1)/w
        return -comp1 -comp2
    #============================================
    # calculates the radius determined by @c, @w using
    # lilley criteria
    #============================================
    def sk_radius(c,w):
        x0 = fmin(lambda x: dtheta_dr(x,c,w), 0,disp=False,xtol=1.0E-10)[0]
        rad = x0 - theta(x0,c,w)/dtheta_dr(x0,c,w)
        return rad
    #================================================================
    # main
    #================================================================
    # 1) Find the minimum of magnetisation, as it is nearly the sk center
    index = np.where( splat.mz == min(splat.mz))[0][0]
    # 2) make a 2d fit to mz using sk_2dprofile function
    start_para = [splat.X[index], splat.Y[index], 2.5, 3.0]
    popt, pcov = curve_fit(sk_2dprofile, [splat.X, splat.Y], splat.mz, start_para, maxfev=2000)
    popt, pcov = curve_fit(sk_2dprofile, [splat.X, splat.Y], splat.mz, popt)
    # the estimated parameters c,w determine the Radius
    radius = sk_radius(popt[2], popt[3])
    return radius, popt, np.sqrt(np.diag(pcov))

#============================================
# function to find the width of a domainwall using curve_fit
#============================================
def dwwidth(splat):
    from scipy.optimize import curve_fit, fmin
    X, Y, mx, my, mz = np.asarray(splat.X), np.asarray(splat.Y), np.asarray(splat.mx), np.asarray(splat.my), np.asarray(splat.mz)
    #============================================
    # 2d fit function for the domainwall profile and a straight line
    #============================================
    def dw_2dprofile(R, r0, phi, w):
        theta = 2.0*np.arctan(np.exp( (R[0]*np.cos(phi) + R[1]*np.sin(phi) -r0)/w))
        return np.cos(theta)
    def dw_2dprofile_reduzed(R, w):
        theta = 2.0*np.arctan(np.exp( (R[0]*np.cos(init_phi) + R[1]*np.sin(init_phi)-init_r0)/w))
        return np.cos(theta)
    def gerade(x, y0, m):
        return y0 + x*m
    #============================================
    # the functions returns the initial values for @phi, the angle between @direction
    # and the x-axis, and also @r0, the distance to the wall on this line
    #============================================
    def get_init_values():
        # find points inside the DW and fit them with a straight line
        index = np.where( abs(mz) < 0.1 )
        popt, pcov = curve_fit(gerade, X[index[0]], Y[index[0]])
        # calculate the initial guess for the angle between @direction and x-axis
        # by using the slope of the fitted straight line
        phi = np.pi/2.0 - np.arctan(abs(popt[1]))
        # to find @r0 set arg=(r-r0)n=0 => r0=rx*cos(phi) + ry*sin(phi)
        # which corresponds to a position on the domainwall
        r0 = splat.X[index[0][0]]*np.cos(phi) + splat.Y[index[0][0]]*np.sin(phi)
        return r0, phi
    #============================================
    # main
    #============================================
    # determine situation: down->up, up->down
    dir = np.sign(splat.mz[0])
    # get initial values for @r0 and @phi and fit DW with them to get @init_w
    init_r0, init_phi = get_init_values()
    popt, pcov = curve_fit(dw_2dprofile_reduzed, [splat.X, splat.Y], np.asarray(splat.mz)*dir, [1.0])
    init_w = popt[0]
    start_para = [init_r0, init_phi, init_w]
    # fit the DW profile to the data using reasonable start parameters
    popt, pcov = curve_fit(dw_2dprofile, [splat.X, splat.Y], np.asarray(splat.mz)*dir, start_para)
    return popt[0], popt[1], popt[2]








#==============================================
# INPUT: initial and final images @init and @final. optionaly the number of images
#        to generate
# OUTPUT: the methods writes the sequence of images to the folder @'geodasic_path'
# DESCRIPTION: This method calculates @steps images on the geodasic
#              path between the initial image @init and the final image
#              @final.
# CONVENTION: the initial image is @image_0 and the final @image_(steps)
#==============================================
def geodasic_path(init, final, steps=10, make_pov=False):
    import os
    from magnetisations import rotate_to_axis, rotation_matrix, rotate_spins, SpinLattice
    from visualizations import write_STM, spinpov
    from shell_commands import make_directory, change_directory

    assert init.size == final.size
    if not os.path.isdir('geodasic_path'):
        make_directory('geodasic_path')

    with change_directory('geodasic_path'):
        # write the inital and final STM files
        write_STM(init, name='image_0.dat')
        write_STM(final, name='image_'+str(steps)+'.dat')
        if make_pov :
            spinpov(init, 'image_0')
            spinpov(final, 'image_'+str(steps))

        # for each image on the geodasic path, calculate the spin configuration
        for n in range(1,steps):
            sl = SpinLattice(size=init.size, key='hex', magmom=init.magmom[0])
            for m in range(len(init.mx)):
                axis, ang = rotate_to_axis([final.mx[m], final.my[m], final.mz[m]], [init.mx[m], init.my[m], init.mz[m]])
                mat = rotation_matrix(axis, ang*float(n)/float(steps))
                tempx, tempy, tempz = rotate_spins([init.mx[m]], [init.my[m]], [init.mz[m]], mat)
                sl.mx[m], sl.my[m], sl.mz[m] = tempx[0], tempy[0], tempz[0]
            tag = 'image_'+str(n)
            write_STM(sl, name=tag+'.dat')
            if make_pov :
                spinpov(sl, tag)



# -*- coding: utf-8 -*-
r"""
Analyzing the results for the two spin precision tests
"""
from tests.test_multilayers.ianalyze import IAnalyzer
from scipy.optimize import curve_fit
import numpy as np
import matplotlib.pyplot as plt
from typing import Callable, Tuple
from pathlib import Path
from python3.shell_commands import writelog


class CAnalyzePrecisionWithoutDamping(IAnalyzer):
    r"""
    Class responsible for analyzing the results of the two spin precision tests
    """

    def __init__(self, logfile: Path, J: float, tempdir: Path, lowerline: int, upperline: int,
                 acceptance_level: float = 1e-2) -> None:
        r"""
        Initializes the analyzer

        Args:
            logfile(Path): log file

            J(float): strength of the interlayer exchange. The type of the interlayer exchange does not affect the
            results of the test.

            tempdir(Path): temporary directory for the calculation

            lowerline(int), upperline(int): index of the line in the STM-file in which the lower spin and the upper spin
            is placed.

            acceptance_level(float): Discrepancy between analytic result and precision frequency determined by fit which
            is allowed to pass the test.
        """
        self._J = J
        self.logfile = logfile
        self.tempdir = tempdir
        self.accept = acceptance_level
        self.lowerline = lowerline
        self.upperline = upperline

    def read_spinevolution(self) -> Tuple[np.ndarray, np.ndarray]:
        r"""
        Reads the time evolution of the upper and the lower spin.

        Returns:
            lower spin and upper spin over time: lowerspin=[[sxt0,syt0,szt0],...]
        """
        lowerspin, upperspin = [], []
        for i in range(100):
            file = open(str(self.tempdir / ('SpinSTM_' + str(i) + '.dat')), 'r')
            ll = 0
            for line in file:
                L = line.split()
                if ll == self.lowerline:
                    lowerspin.append([float(L[3]), float(L[4]), float(L[5])])
                elif ll == self.upperline:
                    upperspin.append([float(L[3]), float(L[4]), float(L[5])])
                else:
                    pass
                ll += 1
        return np.asarray(lowerspin), np.asarray(upperspin)

    def __call__(self) -> bool:
        r"""
        Start analyzing

        Returns:
            all(subtest_results)
        """
        lowerspin, upperspin = self.read_spinevolution()
        # time in ps
        time = np.arange(start=0, stop=100, step=1) * 0.01
        # fit model to upper spin
        subtests = []
        components = ['x', 'y', 'z']
        writelog(self.logfile, message='Checking upper spin precision frequency:...')
        for (idx, c) in enumerate(components):
            writelog(self.logfile, message='      ' + c + '-component:...', end='')
            popt, pcov = curve_fit(self.s_upper(component=c), time, upperspin[:, idx], p0=[self.analytic_frequency()])
            writelog(self.logfile, f'Expected: {self.analytic_frequency()}, Result from Fit: {popt[0]}')
            if abs(popt[0]-self.analytic_frequency()) <= self.accept:
                subtests.append(True)
                writelog(self.logfile, f'      Difference: {abs(popt[0]-self.analytic_frequency())} <'
                                       f' {self.accept}...Passed')
            else:
                subtests.append(False)
                writelog(self.logfile, f'      Difference: {abs(popt[0] - self.analytic_frequency())} >'
                                       f' {self.accept}...Failed!')

        return all(subtests)

        """plt.plot(time, upperspin[:, 0], 'k+')
        plt.plot(time, self.s_upper('x')(t=time, frequency=poptx[0]))
        plt.plot(time, upperspin[:, 1], 'k+')
        plt.plot(time, self.s_upper('y')(t=time, frequency=popty[0]))
        plt.plot(time, upperspin[:, 2], 'k+')
        plt.plot(time, self.s_upper('z')(t=time, frequency=poptz[0]))
        plt.show()"""

    def s_upper(self, component: str) -> Callable:
        r"""
        model functions for upper spin

        Args:
            component(str): say which component of spin is adressed (x, y, z)

        Returns:
            fit function
        """

        def sx(t: np.ndarray, frequency: float) -> np.ndarray:
            r"""
            model function for x component of upper spin

            Args:
                t(np.ndarray): time
                frequency(float): precision frequency

            Returns:
                time evolution of spin component
            """
            return 0.5 * np.cos(t * frequency) + 0.5

        def sy(t: np.ndarray, frequency: float) -> np.ndarray:
            r"""
            model function for y component of upper spin (see sx)
            """
            return (np.sqrt(8) / 4) * np.sin(frequency * t)

        def sz(t: np.ndarray, frequency: float) -> np.ndarray:
            r"""
            model function for z component of upper spin (see sx)
            """
            return -0.5 * np.cos(frequency * t) + 0.5

        if component == 'x':
            return sx
        elif component == 'y':
            return sy
        elif component == 'z':
            return sz

    def analytic_frequency(self) -> float:
        r"""
        Returns:
            the analytic precision frequency. This depends only on the interlayer exchange.
        """
        # constants:
        g = 2.002319304
        hbar = 6.582119569 * 10 ** (-4)  # in eVps
        J = self._J / 1000  # in eV
        return (np.sqrt(8) * g * J) / hbar


